namespace K3D.IO.IFF
{
	using System;
	using System.IO;
	using System.Text;

   using K3D.Geometry;


	/// <summary>
	/// IFFChunk
	/// Represents a single IFF Chunk with a 32 bit size
	/// </summary>
	public class IFFChunk
	{
		// ------------------------------------------------------------------------------
		// Construction

		/// <summary>
		/// Construction of an IFFChunk
		/// </summary>
		public IFFChunk(FileStream stream)
		{
			m_stream = stream;
		}
      

		// ------------------------------------------------------------------------------
		// Public methods

		/// <summary>
		/// Get the IFF Chunk ID
		/// </summary>
		/// <returns>IFF Chunk ID</returns>
		public string GetID()
		{
			if (m_id == null)
			{
				byte[] header = new byte[4];
				m_stream.Read(header, 0, 4);
		      
				m_id = IFFChunk.ReadID(header);
			}
         
			return m_id;
		}
      
		/// <summary>
		/// Skip the remaining bytes in this block
		/// Can be called after GetID() or GetLength()
		/// </summary>
		public void Skip()
		{
			if (m_length == -1)
			{
				GetLength();
			}
   	   
			// skip odd length pad byte
			int iLength = m_length;
			if (iLength % 2 == 1)
			{
				iLength++;
			}
   	   
			m_stream.Seek(iLength, SeekOrigin.Current);
		}
      
		/// <summary>
		/// Get the IFF Chunk Length (4 byte integer)
		/// </summary>
		/// <returns>IFF Chunk Length in bytes</returns>
		public virtual int GetLength()
		{
			if (m_id == null)
			{
				GetID();
			}
         
			if (m_length == -1)
			{
				byte[] length = new byte[4];
				m_stream.Read(length, 0, 4);
		   
				// save length for data lookup
				m_length = IFFChunk.ReadInt32(length);
			}
		   
			return m_length;
		}
      
		/// <summary>
		/// Get the IFF Chunk Data block
		/// </summary>
		/// <returns>IFF Chunk Data block</returns>
		public byte[] GetData()
		{
			if (m_length == -1)
			{
				GetLength();
			}
		   
			byte[] data = new byte[m_length];
			m_stream.Read(data, 0, m_length);
		   
			// skip odd length pad byte
			if (m_length % 2 == 1)
			{
				m_stream.ReadByte();
			}
		   
			return data;
		}
      
      
		// ------------------------------------------------------------------------------
		// Static helpers to read known value types from IFF format files
      
		/// <summary>
		/// Read a 32 bit Integer from raw data in Motorola/Network byte order
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
		/// <returns>32bit integer</returns>
		public static int ReadInt32(byte[] data)
		{
         int theInt = 0;
         for (int i=0; i<4; i++)
         {
            theInt = theInt | (((int)data[i]) << (32-8-(i*8)));
         }
         
         return theInt;
		}

		/// <summary>
		/// Read a 32 bit Integer from raw data in Motorola/Network byte order
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
		/// <param name="offset">Offset into the data (incremented by ref)</param>
		/// <returns>32bit integer</returns>
		public static int ReadInt32(byte[] data, ref int offset)
		{
			int theInt = 0;
			for (int i=0; i<4; i++)
			{
				theInt = theInt | (((int)data[i + offset]) << (32-8-(i*8)));
			}
         
         offset += 4;
         
			return theInt;
		}

      /// <summary>
      /// Read 12 byte (3x4byte) floating point RGB colour triple value
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 12 bytes supplied</param>
      /// <param name="offset">Offset into the data (incremented by ref)</param>
      /// <returns>32bit integerColourTriple</returns>
      public static ColourTriple ReadColourTriple(byte[] data, ref int offset)
      {
         float r,g,b;
         
         r = ReadIEEEfloat32(data, ref offset);
         g = ReadIEEEfloat32(data, ref offset);
         b = ReadIEEEfloat32(data, ref offset);

         offset += 12;
         
         return new ColourTriple(r, g, b);
      }

      /// <summary>
      /// Read 3 byte (padded to 4 bytes) RGB byte colour triple value
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
      /// <param name="offset"></param>
      /// <returns>ColourTriple</returns>
      public static ColourTriple ReadColourTripleByte(byte[] data, ref int offset)
      {
         float r,g,b;

         r = ((float)data[offset+0]) * ColourTriple.RGB_BYTE_TO_FLOAT;
         g = ((float)data[offset+1]) * ColourTriple.RGB_BYTE_TO_FLOAT;
         b = ((float)data[offset+2]) * ColourTriple.RGB_BYTE_TO_FLOAT;

         // skip to even bytes
         offset += 4;

         return new ColourTriple(r, g, b);
      }

		/// <summary>
		/// Reads a variable length index value from the raw data in Motorola/Network byte order
		/// A VX is written as a variable length 2 or 4-byte element. If the index value
		/// is less than 65,280 (0xFF00), then the index is written as an unsigned
		/// two-byte integer. Otherwise the index is written as an unsigned four byte
		/// integer with bits 24-31 set. When reading an index, if the first byte
		/// encountered is 255 (0xFF), then the four-byte form is being used and the
		/// first byte should be discarded or masked out. 
		/// </summary>
		/// <param name="data"></param>
		/// <param name="offset">the offset is incremented by the appropriate 2 or 4 bytes</param>
		/// <returns>The value as an Int32</returns>
		public static int ReadVariableLengthIndex(byte[] data, ref int offset)
		{
			int theInt = 0;

			if ( (byte)data[offset] == (byte)0xFF )
			{
				// 4-byte form
				for (int i=1; i<4; i++)
				{
					theInt = theInt | (((int)data[i + offset]) << (32-8-(i*8)));
				}
				offset += 4;
			}
			else
			{
				// 2-byte form
				for (int i=0; i<2; i++)
				{
					theInt = theInt | (((int)data[i + offset]) << (16-8-(i*8)));
				}
				offset += 2;
			}

			return theInt;
		}

		/// <summary>
		/// Read a 16 bit Integer from raw data in Motorola/Network byte order
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 2 bytes supplied</param>
		/// <returns>16bit integer</returns>
		public static int ReadInt16(byte[] data)
		{
         int theShort = 0;
         for (int i=0; i<2; i++)
         {
            theShort = theShort | (((int)data[i]) << (16-8-(i*8)));
         }
         
         return theShort;
		}

		/// <summary>
		/// Read a 16 bit Integer from raw data in Motorola/Network byte order
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 2 bytes supplied</param>
		/// <param name="offset">Offset into the data (incremented by ref)</param>
		/// <returns>16bit integer</returns>
		public static int ReadInt16(byte[] data, ref int offset)
		{
			int theShort = (((int)data[offset]) << (8));
         theShort = theShort | (((int)data[offset+1]));
         
         offset += 2;
         
			return theShort;
		}
      
		/// <summary>
		/// Read a string value up to the next null byte from raw byte data
		/// </summary>
		/// <param name="data">raw byte data which must contain a null character</param>
		/// <param name="offset">Offset into the data (incremented by ref)</param>
		/// <returns>string value</returns>
		public static string ReadNullTerminatedString(byte[] data, ref int offset)
		{
			StringBuilder buf = new StringBuilder(24);
         
			for (int i=0; i<data.Length; i++)
			{
				if (data[i + offset] == 0)
				{
					// found the null terminator - stop parsing
					break;
				}
            
				buf.Append((char)data[i + offset]);
			}
         
         if (buf.Length != 0)
         {
            offset += (buf.Length + 1);
            // odd length strings cause next string to start on even word boundry
            if (offset % 2 == 1)
            {
               offset++;
            }
         }
         else
         {
            // always skip empty string
            offset += 2;
         }
         
			return buf.ToString();
		}

		/// <summary>
		/// Read an IEEE float value from raw byte data in Motorola/Network byte order
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
		/// <returns>32bit float value</returns>
		public static float ReadIEEEfloat32(byte[] data)
		{
         float ieeefloat;
         
         int se1  = (data[0] + 0x100) % 0x100;
         int e2f1 = (data[1] + 0x100) % 0x100;
         int f2   = (data[2] + 0x100) % 0x100;
         int f3   = (data[3] + 0x100) % 0x100;
         
         double sign = (se1 >> 7 == 0) ? 1.0 : -1.0;
         int exponent = ((se1 & 0x7f) << 1 | (e2f1 >> 7));
         int f1 = e2f1 & 0x7f;
         
         double fraction = ((double)f1) * 0.0078125 + 
            ((double)f2) * 3.0517578125e-05 +
            ((double)f3) * 1.19209289550781e-07;
         
         ieeefloat = (float) (sign * 
            ( Math.Pow(2.0, (double)(exponent - 127)) ) *
            ( 1.0 + fraction ));
         
         return ieeefloat;
		}
      
		/// <summary>
		/// Read an IEEE float value from raw byte data in Motorola/Network byte order
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
		/// <param name="offset">Offset into the data (incremented by ref)</param>
		/// <returns>32bit float value</returns>
		public static float ReadIEEEfloat32(byte[] data, ref int offset)
		{
			float ieeefloat;
         
			int se1  = (data[0 + offset] + 0x100) % 0x100;
			int e2f1 = (data[1 + offset] + 0x100) % 0x100;
			int f2   = (data[2 + offset] + 0x100) % 0x100;
			int f3   = (data[3 + offset] + 0x100) % 0x100;
         
			double sign = (se1 >> 7 == 0) ? 1.0 : -1.0;
			int exponent = ((se1 & 0x7f) << 1 | (e2f1 >> 7));
			int f1 = e2f1 & 0x7f;
         
			double fraction = ((double)f1) * 0.0078125 + 
				((double)f2) * 3.0517578125e-05 +
				((double)f3) * 1.19209289550781e-07;
         
			ieeefloat = (float) (sign * 
				( Math.Pow(2.0, (double)(exponent - 127)) ) *
				( 1.0 + fraction ));
         
         offset += 4;
         
			return ieeefloat;
		}

		/// <summary>
		/// Read a 4 character ASCII ID string from raw byte data
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
		/// <returns>4 characters ASCII ID string</returns>
		public static string ReadID(byte[] data)
		{
         StringBuilder buf = new StringBuilder(4);
         buf.Append((char)data[0]);
         buf.Append((char)data[1]);
         buf.Append((char)data[2]);
         buf.Append((char)data[3]);
		   
         return buf.ToString();
		}

		/// <summary>
		/// Read a 4 character ASCII ID string from raw byte data
		/// </summary>
		/// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
		/// <param name="offset">Offset into the data (incremented by ref)</param>
		/// <returns>4 characters ASCII ID string</returns>
		public static string ReadID(byte[] data, ref int offset)
		{
			StringBuilder buf = new StringBuilder(4);
			buf.Append((char)data[0 + offset]);
			buf.Append((char)data[1 + offset]);
			buf.Append((char)data[2 + offset]);
			buf.Append((char)data[3 + offset]);

         offset += 4;
		   
			return buf.ToString();
		}
      

		// ------------------------------------------------------------------------------
		// Private data
      
		protected FileStream m_stream;
		protected int m_length = -1;
		protected string m_id = null;
      
	} // end class IFFChunk
}