namespace K3D.IO.IFF
{
	using System;
	using System.IO;
	using System.Text;
	using System.Diagnostics;


	/// <summary>
	/// IFFReader
	/// Reads and parses a standard IFF format file
	/// </summary>
	public sealed class IFFReader
	{
		//------------------------------------------------------------------------------
		// Construction
      
		/// <summary>
		/// Construction of an IFFReader reader
		/// </summary>
		public IFFReader(String strFilename)
		{
			m_strFilename = strFilename;
		}
      
      
		//------------------------------------------------------------------------------
		// Public API methods
      
		/// <summary>
		/// Opens the IFF file and returns the file type code
		/// </summary>
		/// <returns>4 characters IFF file type code</returns>
		public string open()
		{
			string strFileType = null;
         
			try
			{
				// open the file
				m_stream = new FileStream(m_strFilename, FileMode.Open, FileAccess.Read);
            
				// check the file is IFF format
				m_iTotalLength = validateHeader();
            
				IFFChunk fileType = new IFFChunk(m_stream);
				strFileType = fileType.GetID();
			}
			catch (IOException ioe)
			{
				close();
				throw new IFFReaderException("I/O Error occured: " + ioe.Message);
			}
			catch (IFFReaderException iffe)
			{
				close();
				throw iffe;
			}
         
			return strFileType;
		}
      
		/// <summary>
		/// Parse an IFF format chunked file visiting each chunk in turn
		/// </summary>
		public void parse(IChunkVisitor visitor)
		{
			if (m_stream == null)
			{
				throw new IFFReaderException("IFF file not open!");
			}
         
			try
			{
				// the default is that we read Chunks until told otherwise to read SubChunks
				bool bIsSubChunk = false;
            
				while (m_stream.Position < (m_iTotalLength - 1))
				{
					// create a new chunk object for the next chunk
					IFFChunk chunk;
					if (bIsSubChunk == false)
					{
						chunk = new IFFChunk(m_stream);
					}
					else
					{
						chunk = new IFFSubChunk(m_stream);
					}
					bIsSubChunk = false;

					// call the visitor object to deal with the chunk as required
					ChunkReturn returnValue = visitor.VisitChunk(chunk, chunk.GetID(), chunk.GetLength());
					switch (returnValue)
					{
						case ChunkReturn.Skip:
							// visitor requests that we skip this chunks data block
							chunk.Skip();
							break;
                  
						case ChunkReturn.ReadNextChunk:
							// visitor request we continue to read chunks into the hierarchy
							bIsSubChunk = false;
							break;

						case ChunkReturn.ReadNextSubChunk:
							// visitor has skipped the chunk data - we can continue
							bIsSubChunk = true;
							break;
					}
				}
            visitor.EndVisit();
			}
			catch (IOException ioe)
			{
				throw new IFFReaderException("I/O Error occured: " + ioe.Message);
			}
			finally
			{
				close();
			}
		}
      
		/// <summary>
		/// Close the IFF file
		/// </summary>
		public void close()
		{
         try
         {
            if (m_stream != null)
            {
               m_stream.Close();
               m_stream = null;
            }
         }
         catch (Exception e)
         {
            Trace.WriteLine("Error during file stream close: " + e.Message);
            Trace.WriteLine(e.StackTrace);
         }
		}
      
      
		//------------------------------------------------------------------------------
		// Private helper methods
      
		/// <summary>
		/// Check the header is valid and return total size of file in bytes
		/// </summary>
		/// <returns>Size of file in bytes</returns>
		private int validateHeader()
		{
			// read the IFF header block
			IFFChunk header = new IFFChunk(m_stream);
			string strHeader = header.GetID();
			int iTotalLength = header.GetLength() + 8;
         
			if (K3DGLForm.DEBUG)
			{
				Trace.WriteLine("Read IFF header chunk: " + strHeader);
				Trace.WriteLine("Total file length: " + iTotalLength);
			}
         
			if (strHeader != "FORM")
			{
				throw new IFFReaderException("File is not in expected format!");
			}
		   
			return iTotalLength;
		}
      
      
		//------------------------------------------------------------------------------
		// Private data
      
		// filename of IFF to parse
		private string m_strFilename = null;

		// stream presenting file being parsed
		private FileStream m_stream = null;

		// total size of file in bytes
		private int m_iTotalLength = 0;
      
	} // end class IFFReader
   

	/// <summary>
	/// IFFReaderException class
	/// </summary>
	public class IFFReaderException : Exception
	{
		public IFFReaderException(string message)
			: base(message)
		{
		}
	}
}