namespace K3D.IO.TDStudio
{
   using System;
   using System.IO;
   using System.Text;

   using K3D.Geometry;


   /// <summary>
   /// Chunk
   /// Represents a single 3DStudio file format Chunk
   /// </summary>
   public sealed class Chunk
   {
      // ------------------------------------------------------------------------------
      // Construction

      /// <summary>
      /// Construction
      /// </summary>
      public Chunk(FileStream stream)
      {
         m_stream = stream;
      }
      

      // ------------------------------------------------------------------------------
      // Public methods

      /// <summary>
      /// Get the 2 byte short Chunk ID
      /// </summary>
      /// <returns>2 byte short Chunk ID</returns>
      public short GetID()
      {
         if (m_id == -1)
         {
            byte[] id = new byte[2];
            m_stream.Read(id, 0, 2);

            m_id = ReadID(id);
         }
         
         return m_id;
      }
      
      /// <summary>
      /// Get the Chunk Length (4 byte integer)
      /// </summary>
      /// <returns>Chunk Length in bytes</returns>
      public int GetLength()
      {
         if (m_id == -1)
         {
            GetID();
         }
         
         if (m_length == -1)
         {
            byte[] length = new byte[4];
            m_stream.Read(length, 0, 4);
		      
            // save length for data lookup
            m_length = Chunk.ReadInt32(length);
         }
		   
         return m_length;
      }
      
      /// <summary>
      /// Get the IFF Chunk Data block
      /// </summary>
      /// <returns>IFF Chunk Data block</returns>
      public byte[] GetData()
      {
         if (m_length == -1)
         {
            GetLength();
         }
		   
         byte[] data = new byte[m_length];
         m_stream.Read(data, 0, m_length);
		   
         // TODO: need this???
         // skip odd length pad byte
         if (m_length % 2 == 1)
         {
            m_stream.ReadByte();
         }
		   
         return data;
      }
      
      
      // ------------------------------------------------------------------------------
      // Static helpers to read known value types from 3DStudio format files
      
      /// <summary>
      /// Read a 32 bit Integer from raw data in Intel byte order
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
      /// <returns>32bit integer</returns>
      public static int ReadInt32(byte[] data)
      {
         int theInt = 0;
         for (int i=0; i<4; i++)
         {
            theInt = theInt | (((int)data[i]) << (i*8));
         }
         
         return theInt;
      }

      /// <summary>
      /// Read a 32 bit Integer from raw data in Intel byte order
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 4 bytes supplied</param>
      /// <param name="offset">Offset into the data (incremented by ref)</param>
      /// <returns>32bit integer</returns>
      public static int ReadInt32(byte[] data, ref int offset)
      {
         int theInt = 0;
         for (int i=0; i<4; i++)
         {
            theInt = theInt | (((int)data[i]) << (i*8));
         }
         
         offset += 4;
         
         return theInt;
      }

      /// <summary>
      /// Read a 16 bit Integer from raw data in Intel byte order
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 2 bytes supplied</param>
      /// <returns>16bit integer</returns>
      public static int ReadInt16(byte[] data)
      {
         int theShort = (int)data[0];
         theShort = theShort | (((int)data[1]) << 8);
         
         return theShort;
      }

      /// <summary>
      /// Read a 16 bit Integer from raw data in Intel byte order
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 2 bytes supplied</param>
      /// <param name="offset">Offset into the data (incremented by ref)</param>
      /// <returns>16bit integer</returns>
      public static int ReadInt16(byte[] data, ref int offset)
      {
         int theShort = (int)data[offset];
         theShort = theShort | (((int)data[offset+1]) << 8);
         
         offset += 2;
         
         return theShort;
      }

      /// <summary>
      /// Read a 2 byte short ID from raw byte data
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 2 bytes supplied</param>
      /// <returns>2 byte short ID</returns>
      public static short ReadID(byte[] data)
      {
         return (short)ReadInt16(data);
      }

      /// <summary>
      /// Read a 2 byte short ID from raw byte data
      /// </summary>
      /// <param name="data">raw byte data with a minimum of 2 bytes supplied</param>
      /// <param name="offset">Offset into the data (incremented by ref)</param>
      /// <returns>2 byte short ID</returns>
      public static short ReadID(byte[] data, ref int offset)
      {
         return (short)ReadInt16(data, ref offset);
      }


      // ------------------------------------------------------------------------------
      // Private data

      /// <summary>
      /// Chunk IDs
      /// </summary>
      public static UInt16 PRIMARY = 0x4D4D;

      // Main Chunks
      public static UInt16 OBJECTINFO = 0x3D3D;			// This gives the version of the mesh and is found right before the material and object information
      public static UInt16 VERSION    = 0x0002;			// This gives the version of the .3ds file

      // sub defines of OBJECTINFO
      public static UInt16 MATERIAL   = 0xAFFF;			// This stored the texture info
      public static UInt16 OBJECT	  = 0x4000;			// This stores the faces, vertices, etc...

      // sub defines of MATERIAL
      public static UInt16 MATNAME    = 0xA000;			// This holds the material name
      public static UInt16 MATDIFFUSE = 0xA020;			// This holds the color of the object/material
      public static UInt16 MATMAP     = 0xA200;			// This is a header for a new material
      public static UInt16 MATMAPFILE = 0xA300;			// This holds the file name of the texture

      public static UInt16 OBJECT_MESH = 0x4100;		// This lets us know that we are reading a new object

      // sub defines of OBJECT_MESH
      public static UInt16 OBJECT_VERTICES   = 0x4110;	// The objects vertices
      public static UInt16 OBJECT_FACES		= 0x4120;	// The objects faces
      public static UInt16 OBJECT_MATERIAL	= 0x4130;	// This is found if the object has a material, either texture map or color
      public static UInt16 OBJECT_UV			= 0x4140;	// The UV texture coordinates
      

      /// <summary>
      /// The filestream this chunk appears in
      /// </summary>
      private FileStream m_stream;

      /// <summary>
      /// Length of the chunk in bytes
      /// </summary>
      private int m_length = -1;

      /// <summary>
      /// 2 byte short ID of the chunk
      /// </summary>
      private short m_id = -1;
      
   } // end class Chunk
}