/*
 * K3D Object Viewer Copyright (C) 1998-2005 Kevin Roast
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 */
namespace K3D
{
	using System;
	using System.IO;
	using System.Drawing;
	using System.Drawing.Imaging;
	using System.Collections;
	using System.ComponentModel;
	using System.Windows.Forms;
	using System.Diagnostics;
	using System.Text;

   using Crownwood.Magic.Menus;

   using K3D.IO.IFF;
   using K3D.ThreeD;
   using K3D.ThreeD.Lightwave;
	using K3D.Services;
	using K3D.Properties;

	using Tao.OpenGl;


	/// <summary>
	/// K3D-GL2 Main Form. This form acts as a controller for the main UI. It is
	/// responsible for setting up the menus, user preferences etc. and then responding
	/// to UI interaction from user.
	/// </summary>
	public sealed class K3DGLForm : System.Windows.Forms.Form
	{
		/// <summary>
		/// Main launching point
		/// </summary>
		[STAThread()]
		public static void Main(string[] args)
		{
			try
			{
				Application.Run(new K3DGLForm(args));
			}
			catch (Exception e)
			{
				MessageBox.Show("Unhandled Error: " + e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
				if (K3DGLForm.DEBUG) Trace.WriteLine("Unhandled Error: " + e.Message);
				if (K3DGLForm.DEBUG) Trace.WriteLine(e.StackTrace);
			}
		}


		// ------------------------------------------------------------------------------
		// Construction
      
		/// <summary>
		/// Constructor
		/// </summary>
		public K3DGLForm(string[] args)
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();
         
			InitWindow();
         InitMenu();
			InitRenderer();
			
			// register file types
			RegistryService.Instance.RegisterFileTypes(Application.ExecutablePath);

         // Load in command line arg file if found
			if (args.Length > 0)
			{
				LoadFile(args[0], true);
			}
		}


		// ------------------------------------------------------------------------------
		// Public API

      /// <summary>
      /// Renderer Instance
      /// </summary>
      public K3DGLRenderer Renderer
      {
         get
         {
            return m_renderer;
         }
      }
		
		/// <summary>
		/// Load and display a 3D object file
		/// </summary>
		/// <param name="file">Full path and file name of valid object file</param>
		public void LoadFile(string file, bool buildFileList)
		{
			LoadFile(file, buildFileList, true);
		}

		/// <summary>
		/// Load and display a 3D object file
		/// </summary>
		/// <param name="file">Full path and file name of valid object file</param>
		/// <param name="resetObject">True to reset the object, false otherwise</param>
		public void LoadFile(string file, bool buildFileList, bool resetObject)
		{
			m_renderer.Render = false;
			
			if (buildFileList == true)
			{
				BuildDirFileList(file);
			}
         
			Application.DoEvents();
			this.Cursor = Cursors.WaitCursor;
			try
			{
				if (File.Exists(file) && Directory.Exists(file) == false)
				{
					IFFReader reader = new IFFReader(file);
					
					string strFileType = reader.open();
					if (K3DGLForm.DEBUG) Trace.WriteLine("Read IFF file type: " + strFileType);
					
					IChunkVisitor visitor = null;
					switch (strFileType)
					{
						case "LWO2":
							visitor = new LWO2Visitor();
							break;
            
						case "LWOB":
							visitor = new LWOBVisitor();
							break;

						//case "TDDD":
							//visitor = new TDDDVisitor();
							//break;
            
						default:
							MessageBox.Show("Unknown file type", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
							break;
					}
					
					// parse the chunks with the appropriate visitor object
					if (visitor != null)
					{
						long starttime, endtime;
						
						// parse the file
						starttime = DateTime.Now.Ticks;

                  this.statusBar.Text = "Loading file... " + MRUFileListService.ShortenPathname(file);
						reader.parse(visitor);
						endtime = DateTime.Now.Ticks;
						if (DEBUG) Trace.WriteLine("Time take to parse file: " + ((float)(endtime - starttime))/10000000.0f + "secs");
						reader = null;
						
						// if we get here without error, the visitor should have created
						// the 3D object, so we can process it now
						starttime = DateTime.Now.Ticks;
                  this.statusBar.Text = "Generating 3D object...";
						m_renderer.Set3DObject(visitor.Get3DObject());
						if (resetObject == true && PropertiesService.Instance.ResetObject == true)
						{
							// reset display variables
							m_renderer.ResetObject();
						}
						endtime = DateTime.Now.Ticks;
						if (DEBUG) Trace.WriteLine("Time take to process 3D data: " + ((float)(endtime - starttime))/10000000.0f + "secs");
						
						// set window title text
						this.Text = "K3D - " + MRUFileListService.ShortenPathname(file);
						
						// set status bar text
						this.statusBar.Text = BuildObjectInfoText(visitor.Get3DObject().ObjectInfo);
						
						// add file to MRU list
						MRUFileListService.Instance.AddFile(file);
						m_lastfile = file;
					}
				}
			}
			catch (IFFReaderException iffe)
			{
				MessageBox.Show("Error reading object file: " + iffe.Message, "Load Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
				if (K3DGLForm.DEBUG) Trace.WriteLine("Error reading object file: " + iffe.Message);
				if (K3DGLForm.DEBUG) Trace.WriteLine(iffe.StackTrace);

				MRUFileListService.Instance.RemoveFile(file);

				this.statusBar.Text = "Error during load of object file.";
			}
			catch (Exception e)
			{
				MessageBox.Show("Internal Error: " + e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
				if (K3DGLForm.DEBUG) Trace.WriteLine("Internal Error: " + e.Message);
				if (K3DGLForm.DEBUG) Trace.WriteLine(e.StackTrace);

				MRUFileListService.Instance.RemoveFile(file);

				this.statusBar.Text = "Error during process of object file.";
			}
			finally
			{
				m_renderer.Render = true;
				m_renderer.Invalidate();
				this.Cursor = Cursors.Default;
			}
		}


		// ------------------------------------------------------------------------------
		// Private helpers
      
		/// <summary>
		/// Build a list of all files in the dir used by the specified file
		/// Also sets the current file index to the specified file
		/// </summary>
		/// <param name="origFileName"></param>
		private void BuildDirFileList(String origFileName)
		{
			string dir = origFileName.Substring(0, origFileName.LastIndexOf('\\') + 1);
			if (Directory.Exists(dir))
			{
				ArrayList filesTemp = new ArrayList(16);

				// recurse down a single directory only
				m_nFileListIndex = 0;
				int count=0;
				string[] dirFiles = Directory.GetFiles(dir, "*.lwo");
				foreach (string dirFile in dirFiles)
				{
					if (File.Exists(dirFile) == true)
					{
						filesTemp.Add(dirFile);
						if (dirFile == origFileName)
						{
							m_nFileListIndex = count;
						}
						count++;
					}
				}
				m_filelist = (string[])filesTemp.ToArray(typeof(string));
			}
			else
			{
				m_filelist = null;
			}
		}

		/// <summary>
		/// Build info text for the specified ObjectInfo structure
		/// </summary>
		/// <param name="info"></param>
		/// <returns></returns>
		private string BuildObjectInfoText(ObjectInfo info)
		{
			StringBuilder buf = new StringBuilder(128);
			buf.Append("Displaying ").Append(info.NumLayers);
			if (info.NumLayers == 1)
			{
				buf.Append(" Layer with ");
			}
			else
			{
				buf.Append(" Layers with ");
			}
			buf.Append(info.NumPoints).Append(" Points, ");
			buf.Append(info.NumPolygons).Append(" Polygons and ");
			buf.Append(info.NumSurfaces).Append(" Surfaces.");

			return buf.ToString();
		}

		/// <summary>
		/// Set the last used window size and position from the registry
		/// </summary>
		private void InitWindow()
		{
			int[] windowValues = RegistryService.Instance.GetIntegerArray(REG_KEY_MAIN, REG_ITEM_WINDOW);
			if (windowValues != null && windowValues.Length == 4)
			{
				this.StartPosition = System.Windows.Forms.FormStartPosition.Manual;
				this.Width = windowValues[0];
				this.Height = windowValues[1];
				this.Top = windowValues[2];
				this.Left = windowValues[3];
			}
			bool maximised = RegistryService.Instance.GetBoolean(REG_KEY_MAIN, REG_ITEM_MAX, false);
			if (maximised == true)
			{
				WindowState = FormWindowState.Maximized;
			}
		}

		/// <summary>
		/// Init the OpenGL object renderer control
		/// </summary>
		private void InitRenderer()
		{
			// OpenGL control will fill the view
			m_renderer = new K3DGLRenderer();
			m_renderer.Dock = DockStyle.Fill;
			m_renderer.Render = false;
			m_renderer.MouseUp += new MouseEventHandler(this.OnMouseUp);
			m_renderer.MouseMove += new MouseEventHandler(this.OnMouseMove);
			m_renderer.MouseDown += new MouseEventHandler(this.OnMouseDown);
			m_renderer.MouseWheel += new MouseEventHandler(this.OnMouseWheel);

         m_renderer.InitializeContexts();
         m_renderer.SetupGLContext();

         // init last used rendering mode
         m_renderer.DisplayMode = (DisplayMode)PropertiesService.Instance.DisplayMode;
         this.toolBar.Buttons[ ((int)m_renderer.DisplayMode) + BUTTON_POINTS ].Pushed = true;
			
			// set the rendering mode toggle toolbar buttons
			this.toolBar.Buttons[ BUTTON_TRANS ].Pushed = PropertiesService.Instance.Transparency;
			this.toolBar.Buttons[ BUTTON_BACKFACES ].Pushed = PropertiesService.Instance.BackFaces;
			
			// light toggle buttons
			InitLightToggleButtons();
			
			this.panel.Controls.Add(m_renderer);
		}
		
		/// <summary>
		/// Init the state of the Light toggle buttons
		/// </summary>
		private void InitLightToggleButtons()
		{
			// light toggle buttons
			LightInfo info = new LightInfo();
			info.DeserialiseFromRegistry(LightsDialog.REG_LIGHT1);
			this.toolBar.Buttons[BUTTON_LIGHT1].Pushed = info.Active;
			info.DeserialiseFromRegistry(LightsDialog.REG_LIGHT2);
			this.toolBar.Buttons[BUTTON_LIGHT2].Pushed = info.Active;
			info.DeserialiseFromRegistry(LightsDialog.REG_LIGHT3);
			this.toolBar.Buttons[BUTTON_LIGHT3].Pushed = info.Active;
			info.DeserialiseFromRegistry(LightsDialog.REG_LIGHT4);
			this.toolBar.Buttons[BUTTON_LIGHT4].Pushed = info.Active;
		}

      /// <summary>
      /// Init the menu control
      /// </summary>
      private void InitMenu()
      {
         // main menu control
         m_menu = new MenuControl();
         m_menu.AnimateStyle = Animation.System;
         m_menu.AnimateTime = 100;
         m_menu.Cursor = System.Windows.Forms.Cursors.Arrow;
         m_menu.Direction = Crownwood.Magic.Common.Direction.Horizontal;
         m_menu.Dock = System.Windows.Forms.DockStyle.Top;
         m_menu.Font = new System.Drawing.Font("Tahoma", 8F, System.Drawing.FontStyle.Regular);
         m_menu.HighlightTextColor = System.Drawing.SystemColors.MenuText;
         m_menu.Name = "main";
         m_menu.Size = new System.Drawing.Size(1600, 22);
         m_menu.Style = Crownwood.Magic.Common.VisualStyle.IDE;
         m_menu.TabIndex = 0;
         m_menu.TabStop = false;
         m_menu.Text = "main";

         // File Menu
         MenuCommand mainMenu = new MenuCommand("&File");
         m_menu.MenuCommands.Add(mainMenu);
         MenuCommand openMenu = new MenuCommand("&Open...", this.imageList, 0, Shortcut.CtrlO, new EventHandler(OnMenuOpen));
         mainMenu.MenuCommands.Add(openMenu);
			MenuCommand saveMenu = new MenuCommand("&Save As Image...", this.imageList, 17, Shortcut.CtrlS, new EventHandler(OnMenuSaveBitmap));
			mainMenu.MenuCommands.Add(saveMenu);
         MenuCommand separator = new MenuCommand("-");
         mainMenu.MenuCommands.Add(separator);

			// inform MRUFileListService of the parent menu to add children too
			MenuCommand mruMenu = new MenuCommand("&Recent Files");
			mainMenu.MenuCommands.Add(mruMenu);
			MRUFileListService.Instance.SetParentMenu(this, mruMenu);

			mainMenu.MenuCommands.Add(separator);
         MenuCommand exitMenu = new MenuCommand("E&xit", new EventHandler(OnMenuExit));
         mainMenu.MenuCommands.Add(exitMenu);

			// Tools Menu
			MenuCommand toolsMenu = new MenuCommand("&Tools");
			m_menu.MenuCommands.Add(toolsMenu);
         MenuCommand resetMenu = new MenuCommand("&Reset Object", this.imageList, 7, new EventHandler(OnMenuResetObject));
         toolsMenu.MenuCommands.Add(resetMenu);
         toolsMenu.MenuCommands.Add(separator);
			MenuCommand optionsMenu = new MenuCommand("&Preferences...", new EventHandler(OnMenuOptions));
			toolsMenu.MenuCommands.Add(optionsMenu);

			// Rendering Menu
			MenuCommand renderMenu = new MenuCommand("&Render");
			m_menu.MenuCommands.Add(renderMenu);
			MenuCommand modesMenu = new MenuCommand("&Modes");
			renderMenu.MenuCommands.Add(modesMenu);
			MenuCommand pointsMenu = new MenuCommand("&Points", this.imageList, 1, Shortcut.CtrlP, new EventHandler(OnMenuPoints));
			modesMenu.MenuCommands.Add(pointsMenu);
			MenuCommand wireMenu = new MenuCommand("&Wireframe", this.imageList, 2, Shortcut.CtrlW, new EventHandler(OnMenuWireframe));
			modesMenu.MenuCommands.Add(wireMenu);
         MenuCommand aawireMenu = new MenuCommand("&Anti-aliased Wireframe", this.imageList, 3, Shortcut.CtrlA, new EventHandler(OnMenuAAWireframe));
         modesMenu.MenuCommands.Add(aawireMenu);
			MenuCommand lineMenu = new MenuCommand("&Hidden Line", this.imageList, 4, Shortcut.CtrlH, new EventHandler(OnMenuHiddenLine));
			modesMenu.MenuCommands.Add(lineMenu);
			MenuCommand flatMenu = new MenuCommand("&Flat Shaded", this.imageList, 5, Shortcut.CtrlF, new EventHandler(OnMenuFlatShaded));
			modesMenu.MenuCommands.Add(flatMenu);
			MenuCommand smoothMenu = new MenuCommand("&Gouraud Shaded", this.imageList, 6, Shortcut.CtrlG, new EventHandler(OnMenuSmoothShaded));
			modesMenu.MenuCommands.Add(smoothMenu);
         renderMenu.MenuCommands.Add(separator);
         MenuCommand transMenu = new MenuCommand("&Transparency", this.imageList, 10, Shortcut.CtrlT, new EventHandler(OnMenuTransparency));
         renderMenu.MenuCommands.Add(transMenu);
         MenuCommand backfaceMenu = new MenuCommand("Show Back &Faces", this.imageList, 8, Shortcut.CtrlF, new EventHandler(OnMenuBackfaces));
         renderMenu.MenuCommands.Add(backfaceMenu);
         MenuCommand bgMenu = new MenuCommand("&Background Color", this.imageList, 9, new EventHandler(OnMenuBackground));
         renderMenu.MenuCommands.Add(bgMenu);

			// Lighting Menu
			MenuCommand lightMenu = new MenuCommand("&Lighting");
			m_menu.MenuCommands.Add(lightMenu);
			MenuCommand ldialogMenu = new MenuCommand("&Light Controls...", this.imageList, 12, Shortcut.CtrlL, new EventHandler(OnMenuLightControls));
			lightMenu.MenuCommands.Add(ldialogMenu);
			lightMenu.MenuCommands.Add(separator);
			MenuCommand light1Menu = new MenuCommand("Toggle Light 1", this.imageList, 13, new EventHandler(OnMenuLight1));
			lightMenu.MenuCommands.Add(light1Menu);
			MenuCommand light2Menu = new MenuCommand("Toggle Light 2", this.imageList, 14, new EventHandler(OnMenuLight2));
			lightMenu.MenuCommands.Add(light2Menu);
			MenuCommand light3Menu = new MenuCommand("Toggle Light 3", this.imageList, 15, new EventHandler(OnMenuLight3));
			lightMenu.MenuCommands.Add(light3Menu);
			MenuCommand light4Menu = new MenuCommand("Toggle Light 4", this.imageList, 16, new EventHandler(OnMenuLight4));
			lightMenu.MenuCommands.Add(light4Menu);
         
         // Help Menu
         MenuCommand helpMenu = new MenuCommand("&Help");
         m_menu.MenuCommands.Add(helpMenu);
         MenuCommand instrMenu = new MenuCommand("&Instructions", new EventHandler(OnMenuInstructions));
         helpMenu.MenuCommands.Add(instrMenu);
         MenuCommand aboutMenu = new MenuCommand("&About K3D", this.imageList, 11, new EventHandler(OnMenuAbout));
         helpMenu.MenuCommands.Add(aboutMenu);
         
         // add to Form
         this.Controls.Add(m_menu);
      }

      /// <summary>
      /// Process a key press in the dialog window.
      /// </summary>
      /// <param name="keyData"></param>
      /// <returns></returns>
      protected override bool ProcessDialogKey(Keys keyData)
      {
         //Trace.WriteLine(keyData.ToString());
         switch (keyData)
         {
            case Keys.Up:
               m_renderer.Rotation.X -= PropertiesService.Instance.KeyRotation;
               m_renderer.Invalidate();
               break;
            case Keys.Down:
               m_renderer.Rotation.X += PropertiesService.Instance.KeyRotation;
               m_renderer.Invalidate();
               break;
            case Keys.Left:
               m_renderer.Rotation.Y -= PropertiesService.Instance.KeyRotation;
               m_renderer.Invalidate();
               break;
            case Keys.Right:
               m_renderer.Rotation.Y += PropertiesService.Instance.KeyRotation;
               m_renderer.Invalidate();
               break;

            case Keys.Oemcomma:
               m_renderer.Scaling /= PropertiesService.Instance.KeyScaling;
               m_renderer.Invalidate();
               break;
            case Keys.OemPeriod:
               m_renderer.Scaling *= PropertiesService.Instance.KeyScaling;
               m_renderer.Invalidate();
               break;

				case Keys.PageUp:
					if (m_filelist != null)
					{
						if (m_nFileListIndex >= 1)
						{
							m_nFileListIndex--;
						}
						else
						{
							m_nFileListIndex = m_filelist.Length -1;
						}
						LoadFile(m_filelist[m_nFileListIndex], false);
					}
					break;
				case Keys.PageDown:
					if (m_filelist != null)
					{
						if (m_nFileListIndex < m_filelist.Length - 1)
						{
							m_nFileListIndex++;
						}
						else
						{
							m_nFileListIndex = 0;
						}
						LoadFile(m_filelist[m_nFileListIndex], false);
					}
					break;

				case Keys.Space:
					m_renderer.ResetObject();
					m_renderer.Invalidate();
					break;
            // TODO: SHIFT+arrows to move object?
         }

         // handle number keys 1-9 & 0 to toggle active layers
         int val = (int)keyData;
         if (val >= 49 && val <= 57)   // D1=49 to D9=57
         {
            m_renderer.ToggleLayer(val - 49);
            m_renderer.Invalidate();
         }
         else if (val == 48)           // D0=48
         {
            m_renderer.ToggleLayer(9);
            m_renderer.Invalidate();
         }

			// TODO: Numeric 1-4 turn lights on/off

         return base.ProcessDialogKey (keyData);
      }


		// ------------------------------------------------------------------------------
		// Menu and Toolbar event handlers

		/// <summary>
		/// Menu Open - load a new object file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
      private void OnMenuOpen(object sender, EventArgs args)
      {
         DialogResult result = this.openFileDialog.ShowDialog();
         if (result == DialogResult.OK)
         {
            LoadFile(this.openFileDialog.FileName, true);
         }
      }

		/// <summary>
		/// Menu Save Bitmap - save the current window as a bitmap file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuSaveBitmap(object sender, EventArgs args)
		{
			if (m_lastfile != null)
			{
				using (Bitmap bitmap = new Bitmap(m_renderer.Width, m_renderer.Height, PixelFormat.Format24bppRgb))
				{
					Rectangle rect = new Rectangle(0, 0, m_renderer.Width, m_renderer.Height);
					BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb);
					Gl.glReadPixels(0, 0, m_renderer.Width, m_renderer.Height, Gl.GL_BGR, Gl.GL_UNSIGNED_BYTE, bitmapData.Scan0);
					bitmap.UnlockBits(bitmapData);
					bitmap.RotateFlip(RotateFlipType.RotateNoneFlipY);
					
					this.saveFileDialog.FileName = m_lastfile.Substring(0, m_lastfile.LastIndexOf('.'));
					if (this.saveFileDialog.ShowDialog() == DialogResult.OK)
					{
						string filename = this.saveFileDialog.FileName;
						string format = filename.Substring(filename.LastIndexOf('.') + 1).ToLower();
						ImageFormat imageFormat;
						switch (format)
						{
							case "jpg":
							case "jpeg":
								imageFormat = ImageFormat.Jpeg;
								break;
							case "png":
								imageFormat = ImageFormat.Png;
								break;
							case "tif":
							case "tiff":
								imageFormat = ImageFormat.Tiff;
								break;
							case "bmp":
							default:
								imageFormat = ImageFormat.Bmp;
								break;
						}
						bitmap.Save(this.saveFileDialog.FileName, imageFormat);
					}
				}
			}
		}

		/// <summary>
		/// Menu Options - Show the options dialog
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuOptions(object sender, EventArgs args)
		{
			PropertiesContainerForm props = new PropertiesContainerForm();
			
			// register the properties pages with the container
			props.AddPropertiesPage(new GeneralProperties());
			props.AddPropertiesPage(new PolygonProperties());
			props.AddPropertiesPage(new DetailProperties());
			
			if (props.ShowDialog() == DialogResult.OK)
			{
				if (props.RefreshLoadedItem == true && m_lastfile != null)
				{
					// properties dialog requested an object refresh
					LoadFile(m_lastfile, false, false);
				}
			}
		}

		/// <summary>
		/// Menu points, change to points display mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuPoints(object sender, EventArgs args)
		{
			ResetDisplayButtons();
			this.toolBar.Buttons[BUTTON_POINTS].Pushed = true;
			m_renderer.DisplayMode = DisplayMode.Points;
			m_renderer.Invalidate();
		}

		/// <summary>
		/// Menu wireframe, change to wireframe display mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuWireframe(object sender, EventArgs args)
		{
			ResetDisplayButtons();
			this.toolBar.Buttons[BUTTON_WIREFRAME].Pushed = true;
			m_renderer.DisplayMode = DisplayMode.Wireframe;
			m_renderer.Invalidate();
		}

      /// <summary>
		/// Menu AA wireframe, change to AA wireframe display mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuAAWireframe(object sender, EventArgs args)
		{
			ResetDisplayButtons();
			this.toolBar.Buttons[BUTTON_AAWIREFRAME].Pushed = true;
			m_renderer.DisplayMode = DisplayMode.AAWireframe;
			m_renderer.Invalidate();
		}

		/// <summary>
		/// Menu hidden line, change to hidden line display mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuHiddenLine(object sender, EventArgs args)
		{
			ResetDisplayButtons();
			this.toolBar.Buttons[BUTTON_HIDDENLINE].Pushed = true;
			m_renderer.DisplayMode = DisplayMode.HiddenLine;
			m_renderer.Invalidate();
		}

		/// <summary>
		/// Menu flat shade, change to flat shaded display mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuFlatShaded(object sender, EventArgs args)
		{
			ResetDisplayButtons();
			this.toolBar.Buttons[BUTTON_FLATSHADE].Pushed = true;
			m_renderer.DisplayMode = DisplayMode.FlatShade;
			m_renderer.Invalidate();
		}

		/// <summary>
		/// Menu smooth shade, change to smooth shaded display mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuSmoothShaded(object sender, EventArgs args)
		{
			ResetDisplayButtons();
			this.toolBar.Buttons[BUTTON_SMOOTHSHADE].Pushed = true;
			m_renderer.DisplayMode = DisplayMode.SmoothShade;
			m_renderer.Invalidate();
		}

      /// <summary>
      /// Menu reset object, reset the position etc. of the object
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="args"></param>
      private void OnMenuResetObject(object sender, EventArgs args)
      {
         m_renderer.ResetObject();
         m_renderer.Invalidate();
      }

      /// <summary>
      /// Menu transparency, toggle polygon transparency
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="args"></param>
      private void OnMenuTransparency(object sender, EventArgs args)
      {
         m_renderer.Transparency = !m_renderer.Transparency;
         this.toolBar.Buttons[BUTTON_TRANS].Pushed = m_renderer.Transparency;
         m_renderer.Invalidate();
      }

      /// <summary>
      /// Menu background colour, set background colour
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="args"></param>
      private void OnMenuBackground(object sender, EventArgs args)
      {
         if (this.colorPicker.ShowDialog() == DialogResult.OK)
         {
            m_renderer.BackgroundColour = this.colorPicker.Color;
            m_renderer.Invalidate();
         }
      }

      /// <summary>
      /// Menu back face display, toggle back face display
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="args"></param>
      private void OnMenuBackfaces(object sender, EventArgs args)
      {
         m_renderer.BackFaces = !m_renderer.BackFaces;
         this.toolBar.Buttons[BUTTON_BACKFACES].Pushed = m_renderer.BackFaces;
         m_renderer.Invalidate();
      }

		/// <summary>
		/// Menu light controls, show the light controls dialog
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuLightControls(object sender, EventArgs args)
		{
			LightsDialog dialog = new LightsDialog();
			dialog.ShowDialog(this);
		
			InitLightToggleButtons();
		}

		/// <summary>
		/// Light 1 toggle button pressed
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuLight1(object sender, EventArgs args)
		{
			ToggleLight(0, BUTTON_LIGHT1);
		}

		/// <summary>
		/// Light 2 toggle button pressed
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuLight2(object sender, EventArgs args)
		{
			ToggleLight(1, BUTTON_LIGHT2);
		}

		/// <summary>
		/// Light 3 toggle button pressed
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuLight3(object sender, EventArgs args)
		{
			ToggleLight(2, BUTTON_LIGHT3);
		}

		/// <summary>
		/// Light 4 toggle button pressed
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuLight4(object sender, EventArgs args)
		{
			ToggleLight(3, BUTTON_LIGHT4);
		}

		/// <summary>
		/// Toggle the specified light on/off
		/// </summary>
		/// <param name="light"></param>
		private void ToggleLight(int lightIndex, int buttonIndex)
		{
			this.toolBar.Buttons[buttonIndex].Pushed = !this.toolBar.Buttons[buttonIndex].Pushed;
			
			LightInfo info = new LightInfo();
			info.DeserialiseFromRegistry(LightsDialog.REG_LIGHT + lightIndex);
			info.Active = this.toolBar.Buttons[buttonIndex].Pushed;
			info.SerialiseToRegistry(LightsDialog.REG_LIGHT + lightIndex);
			
			LightsDialog.UpdateGLLight(Gl.GL_LIGHT0 + lightIndex, info);
			
			m_renderer.Invalidate();
		}

		/// <summary>
		/// Menu instructions, show help dialog
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnMenuInstructions(object sender, EventArgs args)
		{
			MessageBox.Show(MSG_INSTRUCTIONS, "Using K3D...", MessageBoxButtons.OK, MessageBoxIcon.Information);
		}

      /// <summary>
      /// Menu about, show About box
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="args"></param>
      private void OnMenuAbout(object sender, EventArgs args)
      {
         new About().ShowDialog();
      }

		/// <summary>
		/// Menu Exit - quit the app
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
      private void OnMenuExit(object sender, EventArgs args)
      {
			this.Close();
         Application.Exit();
      }

		/// <summary>
		/// Toolbar button clicked
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OnButtonClick(object sender, System.Windows.Forms.ToolBarButtonClickEventArgs e)
		{
			if (e.Button.Tag != null)
			{
            // re-route button press to appropriate menu event handler
				switch( Int32.Parse(e.Button.Tag as string) )
				{
					case BUTTON_POINTS:
						OnMenuPoints(sender, EventArgs.Empty);
						break;
					case BUTTON_WIREFRAME:
						OnMenuWireframe(sender, EventArgs.Empty);
						break;
               case BUTTON_AAWIREFRAME:
                  OnMenuAAWireframe(sender, EventArgs.Empty);
                  break;
					case BUTTON_HIDDENLINE:
						OnMenuHiddenLine(sender, EventArgs.Empty);
						break;
					case BUTTON_FLATSHADE:
						OnMenuFlatShaded(sender, EventArgs.Empty);
						break;
					case BUTTON_SMOOTHSHADE:
						OnMenuSmoothShaded(sender, EventArgs.Empty);
						break;
               case BUTTON_OPENFILE:
                  OnMenuOpen(sender, EventArgs.Empty);
                  break;
               case BUTTON_RESET:
                  OnMenuResetObject(sender, EventArgs.Empty);
                  break;
               case BUTTON_TRANS:
                  OnMenuTransparency(sender, EventArgs.Empty);
                  break;
               case BUTTON_BACKGROUND:
                  OnMenuBackground(sender, EventArgs.Empty);
                  break;
               case BUTTON_BACKFACES:
                  OnMenuBackfaces(sender, EventArgs.Empty);
                  break;
					case BUTTON_LIGHTDIALOG:
						OnMenuLightControls(sender, EventArgs.Empty);
						break;
					case BUTTON_LIGHT1:
						OnMenuLight1(sender, EventArgs.Empty);
						break;
					case BUTTON_LIGHT2:
						OnMenuLight2(sender, EventArgs.Empty);
						break;
					case BUTTON_LIGHT3:
						OnMenuLight3(sender, EventArgs.Empty);
						break;
					case BUTTON_LIGHT4:
						OnMenuLight4(sender, EventArgs.Empty);
						break;
					default:
						break;
				}
			}
		}

		/// <summary>
		/// Reset the pushed state of the mutually exclusive display mode buttons
		/// </summary>
		private void ResetDisplayButtons()
		{
			for (int index=BUTTON_POINTS; index<=BUTTON_SMOOTHSHADE; index++)
			{
				this.toolBar.Buttons[index].Pushed = false;
			}
		}


		// ------------------------------------------------------------------------------
		// Window event handlers

		/// <summary>
		/// Form closing
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OnClosing(object sender, System.ComponentModel.CancelEventArgs e)
		{
			int[] windowValues = new int[4];
			windowValues[0] = this.Width;
			windowValues[1] = this.Height;
			windowValues[2] = this.Top;
			windowValues[3] = this.Left;
			if (this.Top > 0 && this.Left > 0 && this.Width > 0 && this.Height > 0)
			{
				RegistryService.Instance.PutIntegerArray(REG_KEY_MAIN, REG_ITEM_WINDOW, windowValues);
			}
			RegistryService.Instance.PutBoolean(REG_KEY_MAIN, REG_ITEM_MAX, (WindowState == FormWindowState.Maximized));
		}


		// ------------------------------------------------------------------------------
		// Mouse event handlers

		/// <summary>
		/// Mouse button pressed down in main window
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OnMouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				m_bLeftButtonDown = true;
				this.panel.Cursor = Cursors.NoMove2D;
			}
			if (e.Button == MouseButtons.Right)
			{
				m_bRightButtonDown = true;
				this.panel.Cursor = Cursors.NoMove2D;
			}
			m_beginX = e.X;
			m_beginY = e.Y;
		}

		/// <summary>
		/// Mouse moved in main window
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OnMouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (m_bLeftButtonDown == true || m_bRightButtonDown == true)
			{
				// hack to turn the properties value from keyboard scaling of 1-10 into
				// a useful factor for scaling against current window width
				float mouseScale = 5.0f;//_theProperties.mouseScale
				float scaleDivider = (((10.0f - mouseScale))/3.0f) + 0.5f;
				float scaleTemp = (((float)(m_beginX - e.X)) / this.panel.Width) / scaleDivider;

				// similar hack for the keyboard movement factor
				float mouseMove = 5.0f;//_theProperties.mouseMove
				float moveDivider = ((float)(10 - mouseMove)) + 2.5f;
				
				if (m_bRightButtonDown == true)
				{
					// if both buttons pressed, also do a Z axis rotation
					if (m_bLeftButtonDown == true) 
					{
						m_renderer.Rotation.Z += (m_beginX - e.X) / 2.0f;
					}
					// else check for SHIFT with left button
					else 
					{/*
					//theObject.rbuttondown = false;
					// perform Scale if SHIFT key pressed
					/*if ((nFlags & MK_SHIFT) == MK_SHIFT) 
					{
						if ((theObject.beginx - x) < (theViewer.windowwidth/2))
							theObject.scalefactor = theObject.scalefactor / (1.0f + scaletemp);
						else
							theObject.scalefactor = theObject.scalefactor * (1.0f + scaletemp);
						theObject.beginx = x;
						theObject.beginy = y;
					}
						// else rotate in Z and X axis
					else 
					{*/
						m_renderer.Rotation.X -= (m_beginY - e.Y)/2.0f;
						m_renderer.Rotation.Y -= (m_beginX - e.X)/2.0f;
					}
				}
				else if (m_bLeftButtonDown == true) 
				{
					// if left button pressed, translate the object
					if ((m_beginY - e.Y) < (m_renderer.Height / 2))
					{
						m_renderer.Position.Y += (m_beginY - e.Y) / moveDivider;
					}
					else
					{
						m_renderer.Position.Y -= (m_beginY - e.Y) / moveDivider;
					}

					if ((m_beginX - e.X) < (m_renderer.Width / 2))
					{
						m_renderer.Position.X -= (m_beginX - e.X) / moveDivider;
					}
					else
					{
						m_renderer.Position.X += (m_beginX - e.X) / moveDivider;
					}
				}

				m_beginX = e.X;
				m_beginY = e.Y;
				
				m_renderer.Invalidate();
			}
		}

		/// <summary>
		/// Mouse button released in main window
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OnMouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				m_bLeftButtonDown = false;
			}
			if (e.Button == MouseButtons.Right)
			{
				m_bRightButtonDown = false;
			}
			if (m_bLeftButtonDown == false && m_bRightButtonDown == false)
			{
				this.panel.Cursor = Cursors.Default;
			}
		}

		/// <summary>
		/// User rolled the mouse wheel - scale the object
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OnMouseWheel(object sender, MouseEventArgs e)
		{
			if (e.Delta > 0)
			{
				m_renderer.Scaling *= (((float)e.Delta) * 0.01f);
			}
			else
			{
				m_renderer.Scaling /= (((float)0 - e.Delta) * 0.01f);
			}
			m_renderer.Invalidate();
		}


		// ------------------------------------------------------------------------------
		// Drag & Drop event handlers

      /// <summary>
      /// Item(s) dragged over window.
      /// We allow "FileDrop" type to appear as a dropable item
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="e"></param>
		private void OnDragOver(object sender, System.Windows.Forms.DragEventArgs e)
		{
			if (e.Data.GetDataPresent(DataFormats.FileDrop) == true) 
			{
				e.Effect = DragDropEffects.Copy;
			}
			else
			{
				e.Effect = DragDropEffects.None;
			}
		}

      /// <summary>
      /// User has dropped a valid dragdrop item package
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="e"></param>
		private void OnDragDrop(object sender, System.Windows.Forms.DragEventArgs e)
		{
         string[] files = e.Data.GetData(DataFormats.FileDrop) as string[];
         if (files != null)
         {
            ArrayList filesTemp = new ArrayList(16);

            // check each file in the drop list
            foreach (string file in files)
            {
               if (File.Exists(file) == true && Directory.Exists(file) == false)
               {
                  filesTemp.Add(file);
               }
               else if (Directory.Exists(file) == true)
               {
                  // recurse down a single directory only
                  string[] dirFiles = Directory.GetFiles(file, "*.lwo");
                  foreach (string dirFile in dirFiles)
                  {
                     if (File.Exists(dirFile) == true)
                     {
                        filesTemp.Add(dirFile);
                     }
                  }
               }
            }

				m_filelist = (string[])filesTemp.ToArray(typeof(string));

				m_nFileListIndex = 0;

				// import the first file in the drag/drop list
				LoadFile(m_filelist[m_nFileListIndex], false);
         }
		}


		// ------------------------------------------------------------------------------
		// Designer code

		#region Windows Form Designer generated code
      /// <summary>
      /// Clean up any resources being used.
      /// </summary>
      protected override void Dispose( bool disposing )
      {
         if( disposing )
         {
            if(components != null)
            {
               this.m_renderer.DestroyContexts();
               components.Dispose();
            }
         }
         base.Dispose( disposing );
      }
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(K3DGLForm));
			this.statusBar = new System.Windows.Forms.StatusBar();
			this.panel = new System.Windows.Forms.Panel();
			this.openFileDialog = new System.Windows.Forms.OpenFileDialog();
			this.toolBar = new System.Windows.Forms.ToolBar();
			this.sep0 = new System.Windows.Forms.ToolBarButton();
			this.openButton = new System.Windows.Forms.ToolBarButton();
			this.sep1 = new System.Windows.Forms.ToolBarButton();
			this.resetButton = new System.Windows.Forms.ToolBarButton();
			this.sep2 = new System.Windows.Forms.ToolBarButton();
			this.pointsButton = new System.Windows.Forms.ToolBarButton();
			this.wireframeButton = new System.Windows.Forms.ToolBarButton();
			this.aawireframeButton = new System.Windows.Forms.ToolBarButton();
			this.hiddenlineButton = new System.Windows.Forms.ToolBarButton();
			this.flatshadeButton = new System.Windows.Forms.ToolBarButton();
			this.smoothshadeButton = new System.Windows.Forms.ToolBarButton();
			this.sep3 = new System.Windows.Forms.ToolBarButton();
			this.transButton = new System.Windows.Forms.ToolBarButton();
			this.backfaceButton = new System.Windows.Forms.ToolBarButton();
			this.backgroundButton = new System.Windows.Forms.ToolBarButton();
			this.sep4 = new System.Windows.Forms.ToolBarButton();
			this.lightsButton = new System.Windows.Forms.ToolBarButton();
			this.light1Button = new System.Windows.Forms.ToolBarButton();
			this.light2Button = new System.Windows.Forms.ToolBarButton();
			this.light3Button = new System.Windows.Forms.ToolBarButton();
			this.light4Button = new System.Windows.Forms.ToolBarButton();
			this.imageList = new System.Windows.Forms.ImageList(this.components);
			this.colorPicker = new System.Windows.Forms.ColorDialog();
			this.saveFileDialog = new System.Windows.Forms.SaveFileDialog();
			this.SuspendLayout();
			// 
			// statusBar
			// 
			this.statusBar.Location = new System.Drawing.Point(0, 371);
			this.statusBar.Name = "statusBar";
			this.statusBar.Size = new System.Drawing.Size(408, 20);
			this.statusBar.TabIndex = 0;
			this.statusBar.Text = "Ready.";
			// 
			// panel
			// 
			this.panel.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
				| System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.panel.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D;
			this.panel.Location = new System.Drawing.Point(0, 50);
			this.panel.Name = "panel";
			this.panel.Size = new System.Drawing.Size(408, 320);
			this.panel.TabIndex = 1;
			// 
			// openFileDialog
			// 
			this.openFileDialog.AddExtension = false;
			this.openFileDialog.DefaultExt = "lwo";
			this.openFileDialog.Filter = "Lightwave objects (*.lwo)|*.lwo|All files (*.*)|*.*";
			this.openFileDialog.Title = "Select Object File...";
			// 
			// toolBar
			// 
			this.toolBar.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.toolBar.Buttons.AddRange(new System.Windows.Forms.ToolBarButton[] {
																											  this.sep0,
																											  this.openButton,
																											  this.sep1,
																											  this.resetButton,
																											  this.sep2,
																											  this.pointsButton,
																											  this.wireframeButton,
																											  this.aawireframeButton,
																											  this.hiddenlineButton,
																											  this.flatshadeButton,
																											  this.smoothshadeButton,
																											  this.sep3,
																											  this.transButton,
																											  this.backfaceButton,
																											  this.backgroundButton,
																											  this.sep4,
																											  this.lightsButton,
																											  this.light1Button,
																											  this.light2Button,
																											  this.light3Button,
																											  this.light4Button});
			this.toolBar.ButtonSize = new System.Drawing.Size(20, 20);
			this.toolBar.Dock = System.Windows.Forms.DockStyle.None;
			this.toolBar.DropDownArrows = true;
			this.toolBar.ImageList = this.imageList;
			this.toolBar.Location = new System.Drawing.Point(0, 24);
			this.toolBar.Name = "toolBar";
			this.toolBar.ShowToolTips = true;
			this.toolBar.Size = new System.Drawing.Size(409, 28);
			this.toolBar.TabIndex = 2;
			this.toolBar.ButtonClick += new System.Windows.Forms.ToolBarButtonClickEventHandler(this.OnButtonClick);
			// 
			// sep0
			// 
			this.sep0.Style = System.Windows.Forms.ToolBarButtonStyle.Separator;
			// 
			// openButton
			// 
			this.openButton.ImageIndex = 0;
			this.openButton.Tag = "1";
			this.openButton.ToolTipText = "Open File...";
			// 
			// sep1
			// 
			this.sep1.Style = System.Windows.Forms.ToolBarButtonStyle.Separator;
			// 
			// resetButton
			// 
			this.resetButton.ImageIndex = 7;
			this.resetButton.Tag = "3";
			this.resetButton.ToolTipText = "Reset object settings";
			// 
			// sep2
			// 
			this.sep2.Style = System.Windows.Forms.ToolBarButtonStyle.Separator;
			// 
			// pointsButton
			// 
			this.pointsButton.ImageIndex = 1;
			this.pointsButton.Tag = "5";
			this.pointsButton.ToolTipText = "Display object points";
			// 
			// wireframeButton
			// 
			this.wireframeButton.ImageIndex = 2;
			this.wireframeButton.Tag = "6";
			this.wireframeButton.ToolTipText = "Display object as a wireframe";
			// 
			// aawireframeButton
			// 
			this.aawireframeButton.ImageIndex = 3;
			this.aawireframeButton.Tag = "7";
			this.aawireframeButton.ToolTipText = "Display object as anti-aliased wireframe";
			// 
			// hiddenlineButton
			// 
			this.hiddenlineButton.ImageIndex = 4;
			this.hiddenlineButton.Tag = "8";
			this.hiddenlineButton.ToolTipText = "Display object as hidden-line wireframe";
			// 
			// flatshadeButton
			// 
			this.flatshadeButton.ImageIndex = 5;
			this.flatshadeButton.Tag = "9";
			this.flatshadeButton.ToolTipText = "Display object with flat shading";
			// 
			// smoothshadeButton
			// 
			this.smoothshadeButton.ImageIndex = 6;
			this.smoothshadeButton.Tag = "10";
			this.smoothshadeButton.ToolTipText = "Display object with smooth shading";
			// 
			// sep3
			// 
			this.sep3.Style = System.Windows.Forms.ToolBarButtonStyle.Separator;
			// 
			// transButton
			// 
			this.transButton.ImageIndex = 10;
			this.transButton.Tag = "12";
			this.transButton.ToolTipText = "Enable or disable transparency";
			// 
			// backfaceButton
			// 
			this.backfaceButton.ImageIndex = 8;
			this.backfaceButton.Tag = "13";
			this.backfaceButton.ToolTipText = "Enable or disable back face display";
			// 
			// backgroundButton
			// 
			this.backgroundButton.ImageIndex = 9;
			this.backgroundButton.Tag = "14";
			this.backgroundButton.ToolTipText = "Change background colour";
			// 
			// sep4
			// 
			this.sep4.Style = System.Windows.Forms.ToolBarButtonStyle.Separator;
			// 
			// lightsButton
			// 
			this.lightsButton.ImageIndex = 12;
			this.lightsButton.Tag = "16";
			this.lightsButton.ToolTipText = "Show the Light Controls dialog";
			// 
			// light1Button
			// 
			this.light1Button.ImageIndex = 13;
			this.light1Button.Pushed = true;
			this.light1Button.Tag = "17";
			this.light1Button.ToolTipText = "Toggle Light 1";
			// 
			// light2Button
			// 
			this.light2Button.ImageIndex = 14;
			this.light2Button.Tag = "18";
			this.light2Button.ToolTipText = "Toggle Light 2";
			// 
			// light3Button
			// 
			this.light3Button.ImageIndex = 15;
			this.light3Button.Tag = "19";
			this.light3Button.ToolTipText = "Toggle Light 3";
			// 
			// light4Button
			// 
			this.light4Button.ImageIndex = 16;
			this.light4Button.Tag = "20";
			this.light4Button.ToolTipText = "Toggle Light 4";
			// 
			// imageList
			// 
			this.imageList.ColorDepth = System.Windows.Forms.ColorDepth.Depth32Bit;
			this.imageList.ImageSize = new System.Drawing.Size(16, 16);
			this.imageList.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("imageList.ImageStream")));
			this.imageList.TransparentColor = System.Drawing.Color.Transparent;
			// 
			// colorPicker
			// 
			this.colorPicker.AnyColor = true;
			this.colorPicker.Color = System.Drawing.Color.DarkGray;
			this.colorPicker.FullOpen = true;
			// 
			// saveFileDialog
			// 
			this.saveFileDialog.Filter = "Jpeg Files (*.jpg)|*.jpg|Bitmap Files (*.bmp)|*.bmp|PNG Files (*.png)|*.png|Tiff " +
				"Files (*.tif)|*.tif";
			this.saveFileDialog.Title = "Save Image File...";
			// 
			// K3DGLForm
			// 
			this.AllowDrop = true;
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(408, 391);
			this.Controls.Add(this.panel);
			this.Controls.Add(this.toolBar);
			this.Controls.Add(this.statusBar);
			this.Cursor = System.Windows.Forms.Cursors.Arrow;
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.MinimumSize = new System.Drawing.Size(300, 400);
			this.Name = "K3DGLForm";
			this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
			this.Text = "K3D";
			this.Closing += new System.ComponentModel.CancelEventHandler(this.OnClosing);
			this.DragOver += new System.Windows.Forms.DragEventHandler(this.OnDragOver);
			this.DragDrop += new System.Windows.Forms.DragEventHandler(this.OnDragDrop);
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// Controls
		/// </summary>
		private System.ComponentModel.IContainer components;
		private System.Windows.Forms.StatusBar statusBar;
		private System.Windows.Forms.Panel panel;
      private System.Windows.Forms.OpenFileDialog openFileDialog;
      private System.Windows.Forms.ToolBar toolBar;
		private System.Windows.Forms.ImageList imageList;
		private System.Windows.Forms.ToolBarButton openButton;
		private System.Windows.Forms.ToolBarButton sep1;
		private System.Windows.Forms.ToolBarButton pointsButton;
		private System.Windows.Forms.ToolBarButton wireframeButton;
		private System.Windows.Forms.ToolBarButton sep0;
		private System.Windows.Forms.ToolBarButton sep2;
		private System.Windows.Forms.ToolBarButton resetButton;
		private System.Windows.Forms.ToolBarButton hiddenlineButton;
		private System.Windows.Forms.ToolBarButton flatshadeButton;
		private System.Windows.Forms.ToolBarButton smoothshadeButton;
		private System.Windows.Forms.ToolBarButton sep3;
		private System.Windows.Forms.ToolBarButton transButton;
		private System.Windows.Forms.ToolBarButton backfaceButton;
		private System.Windows.Forms.ToolBarButton backgroundButton;
      private System.Windows.Forms.ToolBarButton aawireframeButton;
		private System.Windows.Forms.ToolBarButton sep4;
		private System.Windows.Forms.ToolBarButton lightsButton;
		private System.Windows.Forms.ToolBarButton light1Button;
		private System.Windows.Forms.ToolBarButton light2Button;
		private System.Windows.Forms.ToolBarButton light3Button;
		private System.Windows.Forms.ToolBarButton light4Button;
      private System.Windows.Forms.ColorDialog colorPicker;
		private System.Windows.Forms.SaveFileDialog saveFileDialog;


		// ------------------------------------------------------------------------------
		// Private data

		public static bool DEBUG = false;
		
      /// <summary>
      /// Help text
      /// </summary>
      private const string MSG_INSTRUCTIONS =
         "Welcome to the K3D object viewer!\r\n\r\n" +
         "Use the mouse or keyboard to control the object view.\r\n" +
         "The left button moves the object and the right button rotates the object.\r\n" +
         "The mouse wheel can be used to scale the object in and out of the screen.\r\n" +
         "The arrow keys rotate the object, the '<' and '>' keys scale the object.\r\n" +
         "The number keys from 1-9 can be used to toggle layer visiblity on multi-layers objects.\r\n" +
         "Press SPACE to reset the object in the view.\r\n\r\n" + 
         "Double click a .LWO object to load it into K3D.\r\n" +
         "Drag and drop a list of files and folders from Explorer onto the K3D window to view multiple objects.\r\n" +
         "Use the Page Up and Page Down keys to view the objects in the list.\r\n\r\n" +
         "Enjoy! :)";

		/// <summary>
		/// Registry key/item names
		/// </summary>
		private const string REG_KEY_MAIN		= "Main";
		private const string REG_ITEM_WINDOW	= "Window";
		private const string REG_ITEM_MAX      = "Maximised";

		/// <summary>
		/// Indexes to toolbar buttons
		/// </summary>
		private const int BUTTON_OPENFILE		= 1;
		private const int BUTTON_RESET			= 3;
		private const int BUTTON_POINTS			= 5;
		private const int BUTTON_WIREFRAME		= 6;
      private const int BUTTON_AAWIREFRAME   = 7;
		private const int BUTTON_HIDDENLINE		= 8;
		private const int BUTTON_FLATSHADE		= 9;
		private const int BUTTON_SMOOTHSHADE	= 10;
		private const int BUTTON_TRANS			= 12;
		private const int BUTTON_BACKFACES  	= 13;
		private const int BUTTON_BACKGROUND		= 14;
		private const int BUTTON_LIGHTDIALOG	= 16;
		private const int BUTTON_LIGHT1			= 17;
		private const int BUTTON_LIGHT2			= 18;
		private const int BUTTON_LIGHT3			= 19;
		private const int BUTTON_LIGHT4			= 20;

      /// <summary>
      /// The XP look menu control
      /// </summary>
      private MenuControl m_menu = null;

      /// <summary>
      /// List of files from a drag-drop or file browse operation
      /// </summary>
      private string[] m_filelist = null;

		/// <summary>
		/// Current index into the file list
		/// </summary>
		private int m_nFileListIndex = 0;

		/// <summary>
		/// Last file loaded
		/// </summary>
		private string m_lastfile = null;

      /// <summary>
      /// The OpenGL renderer
      /// </summary>
      private K3DGLRenderer m_renderer = null;

		/// <summary>
		/// Whether the user is trying to drag object in the main window
		/// </summary>
		private bool m_bLeftButtonDown = false;
		private bool m_bRightButtonDown = false;

		/// <summary>
		/// Initial mouse positions when user is dragging the object
		/// </summary>
		private int m_beginX = 0;
		private int m_beginY = 0;
	}
}
