/*
 * K3D Object Viewer Copyright (C) 1998-2005 Kevin Roast
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 */
namespace K3D
{
	using System;
	using System.Drawing;
	using System.Text;
	using System.Windows.Forms;
   using System.Diagnostics;
   using System.Collections;

	using Tao.OpenGl;
	using Tao.Platform.Windows;

   using K3D.Geometry;
   using K3D.ThreeD;
	using K3D.Services;


	/// <summary>
	/// K3DGLRenderer
	/// K3D OpenGL Rendering entry point
	/// </summary>
	public sealed class K3DGLRenderer : K3DGLControl
	{
      // ------------------------------------------------------------------------------
      // Control overrides

      /// <summary>
      /// Init the GL context
      /// </summary>
		public void SetupGLContext()
		{
         if (K3DGLForm.DEBUG) Trace.WriteLine("K3DGLRenderer::InitGLContext()");
      	
         // Rendering mode
			//Gl.glEnable(Gl.GL_TEXTURE_2D);                            // Enable Texture Mapping
			Gl.glShadeModel(Gl.GL_SMOOTH);                              // Enable Smooth Shading
         // get background colour from Properties
         System.Drawing.Color bg = BackgroundColour;
			Gl.glClearColor( ((float)bg.R)*ColourTriple.RGB_BYTE_TO_FLOAT,
                          ((float)bg.G)*ColourTriple.RGB_BYTE_TO_FLOAT,
                          ((float)bg.B)*ColourTriple.RGB_BYTE_TO_FLOAT, 0.0f );
			Gl.glClearDepth(1.0f);                                      // Depth Buffer Setup
         Gl.glEnable(Gl.GL_DEPTH_TEST);                              // Enables Depth Testing
         
         if (PropertiesService.Instance.BackFaces == true)
         {
            Gl.glDisable(Gl.GL_CULL_FACE);
         }
         else
         {
            Gl.glEnable(Gl.GL_CULL_FACE);
         }
			Gl.glLightModeli(Gl.GL_LIGHT_MODEL_TWO_SIDE, (int)Gl.GL_FALSE);	// For LW single sided polys by default
         
			Gl.glDepthFunc(Gl.GL_LEQUAL);                               // The Type Of Depth Testing
			Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);			// rendering mode
         
         // Rendering hints
			Gl.glHint(Gl.GL_PERSPECTIVE_CORRECTION_HINT, Gl.GL_NICEST); // Nice Perspective Calculations
         Gl.glHint(Gl.GL_POINT_SMOOTH_HINT, Gl.GL_NICEST);
         Gl.glHint(Gl.GL_LINE_SMOOTH_HINT, Gl.GL_NICEST);
         
         Gl.glEnable(Gl.GL_POINT_SMOOTH);
         Gl.glDisable(Gl.GL_LINE_SMOOTH);                            // only for AA wireframe mode
         
         // Lighting
         InitLights();
			Gl.glEnable(Gl.GL_LIGHTING);
			Gl.glEnable(Gl.GL_COLOR_MATERIAL);	// track the current colour as a material
			Gl.glEnable(Gl.GL_NORMALIZE);
         
			// Transparency and blending
			if (PropertiesService.Instance.Transparency == true || DisplayMode == DisplayMode.AAWireframe)
			{
				Gl.glEnable(Gl.GL_BLEND);
			}
			else
			{
		      Gl.glDisable(Gl.GL_BLEND);
			}
			Gl.glBlendFunc(Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA);
		}
      
      /// <summary>
      /// GL Draw render method
      /// </summary>
		public override void GLDraw()
		{
         //if (K3DGLForm.DEBUG) Trace.WriteLine("K3DGLRenderer::GLDraw()");
			
			// Clear The Screen And The Depth Buffer
			Gl.glClear(Gl.GL_COLOR_BUFFER_BIT | Gl.GL_DEPTH_BUFFER_BIT);
			
         if (m_bRender == false || m_3dObject == null)
         {
            return;
         }
			
         // rotate, scale and translate object
         Gl.glMatrixMode(Gl.GL_MODELVIEW);
         Gl.glLoadIdentity();
         Gl.glPushMatrix();
			
         Gl.glTranslated(-m_position.X, m_position.Y, 0);	// LW6?
         Gl.glScalef(m_scale, m_scale, m_scale);
         Gl.glRotatef(m_rotation.X, 1.0f, 0.0f, 0.0f);
         Gl.glRotatef(-m_rotation.Y, 0.0f, 1.0f, 0.0f);		// LW6?
         Gl.glRotatef(m_rotation.Z, 0.0f, 0.0f, 1.0f);
			
         // render OpenGL display lists
         // check for hidden line display mode here (special case for rendering)
         if (m_displayMode == DisplayMode.HiddenLine) 
         {
            Gl.glColorMask(Gl.GL_FALSE, Gl.GL_FALSE, Gl.GL_FALSE, Gl.GL_FALSE);
            Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);
            CallOpenGLLists();
            Gl.glColorMask(Gl.GL_TRUE, Gl.GL_TRUE, Gl.GL_TRUE, Gl.GL_TRUE);
            Gl.glLineWidth(1.5f);
            Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_LINE);
            CallOpenGLLists();
            Gl.glLineWidth(1.0f);
         }
         else 
         {
            CallOpenGLLists();
         }

         Gl.glPopMatrix();
         Gl.glFlush();
		}

      /// <summary>
      /// Call all available OpenGL display lists
      /// </summary>
      private void CallOpenGLLists()
      {
         ArrayList listLayers = new ArrayList(m_3dObject.OpenGLListLength);
         for (int i=0; i<m_3dObject.OpenGLListLength; i++)
         {
            if ( ((bool)m_activeLayers[i]) == true )
            {
               listLayers.Add(i + 1);
            }
         }
         int[] lists = (int[])listLayers.ToArray(typeof(int));
         Gl.glCallLists(lists.Length, Gl.GL_INT, lists);
      }

      /// <summary>
      /// Initialise the OpenGL lights
      /// </summary>
      private void InitLights()
      {
         LightInfo lightInfo = new LightInfo();
         
         // LIGHT 1
         if (lightInfo.DeserialiseFromRegistry(LightsDialog.REG_LIGHT1) == false)
         {
            // set defaults for Light1
            lightInfo.Colour    = Light1Colour;
            lightInfo.Intensity = Light1Intensity;
            lightInfo.Specular  = Light1Specular;
            lightInfo.Position  = Light1Position;
            lightInfo.Active    = true;
            
            // save light details for first time
            lightInfo.SerialiseToRegistry(LightsDialog.REG_LIGHT1);
         }
         LightsDialog.UpdateGLLight(Gl.GL_LIGHT0, lightInfo);

         // LIGHT 2
         if (lightInfo.DeserialiseFromRegistry(LightsDialog.REG_LIGHT2) == false)
         {
            // set defaults for Light2
            lightInfo.Colour    = Light2Colour;
            lightInfo.Intensity = Light2Intensity;
            lightInfo.Specular  = Light2Specular;
            lightInfo.Position  = Light2Position;
            lightInfo.Active    = true;
            
            // save light details for first time
            lightInfo.SerialiseToRegistry(LightsDialog.REG_LIGHT2);
         }
         LightsDialog.UpdateGLLight(Gl.GL_LIGHT1, lightInfo);

         // LIGHT 3
         if (lightInfo.DeserialiseFromRegistry(LightsDialog.REG_LIGHT3) == false)
         {
            // set defaults for Light3
            lightInfo.Colour    = Light3Colour;
            lightInfo.Intensity = Light3Intensity;
            lightInfo.Specular  = Light3Specular;
            lightInfo.Position  = Light3Position;
            lightInfo.Active    = false;
            
            // save light details for first time
            lightInfo.SerialiseToRegistry(LightsDialog.REG_LIGHT3);
         }
         LightsDialog.UpdateGLLight(Gl.GL_LIGHT2, lightInfo);

         // LIGHT 4
         if (lightInfo.DeserialiseFromRegistry(LightsDialog.REG_LIGHT4) == false)
         {
            // set defaults for Light4
            lightInfo.Colour    = Light4Colour;
            lightInfo.Intensity = Light4Intensity;
            lightInfo.Specular  = Light4Specular;
            lightInfo.Position  = Light4Position;
            lightInfo.Active    = false;
            
            // save light details for first time
            lightInfo.SerialiseToRegistry(LightsDialog.REG_LIGHT4);
         }
         LightsDialog.UpdateGLLight(Gl.GL_LIGHT3, lightInfo);
      }


      // ------------------------------------------------------------------------------
      // Public API

      /// <summary>
      /// Whether the control should render the OpenGL data or not
      /// </summary>
      public bool Render
      {
         get
         {
            return m_bRender;
         }
         set
         {
            m_bRender = value;
         }
      }

      /// <summary>
      /// Set a new 3D object data to display
      /// </summary>
      /// <param name="three3dObject">ThreeDObject</param>
      public void Set3DObject(ThreeDObject tdObject)
      {
         // kill old object and OpenGL display lists if present
         if (m_3dObject != null)
         {
            Gl.glDeleteLists(1, m_3dObject.OpenGLListLength);
            m_3dObject = null;
				System.GC.Collect(1);
         }
         
         m_3dObject = tdObject;
         
         // prepare the object
         m_3dObject.CentreAndScale();
         m_3dObject.CalculateFacetNormals();
         m_3dObject.CalculateVertexNormals();
         m_3dObject.BuildOpenGLDisplayLists();
         if (K3DGLForm.DEBUG) Trace.WriteLine(m_3dObject.ToString());
         m_3dObject.FreeGeometry();
			
			ResetActiveLayers();
      }

      /// <summary>
      /// Toggle visibility for a layer
      /// </summary>
      /// <param name="layerIndex"></param>
      public void ToggleLayer(int layerIndex)
      {
         if (layerIndex < m_activeLayers.Count)
         {
            m_activeLayers[layerIndex] = !(bool)m_activeLayers[layerIndex];
         }
      }

		/// <summary>
		/// Reset rotation, position and scale for the current object
		/// </summary>
		public void ResetObject()
		{
			m_rotation.Y = 180.0f;
			m_rotation.X = m_rotation.Z = 0.0f;
			m_position.X = m_position.Y = m_position.Z = 0.0f;
			m_scale = 1.0f;

			ResetActiveLayers();
		}

		/// <summary>
		// Reset layer on/off list
		/// </summary>
		public void ResetActiveLayers()
		{
			m_activeLayers.Clear();
			if (m_3dObject != null)
			{
				for (int layer=0; layer<m_3dObject.OpenGLListLength; layer++)
				{
					m_activeLayers.Add(true);
				}
			}
		}

		/// <summary>
		/// Rotation in degrees
		/// </summary>
		public K3D.Geometry.Vector Rotation
		{
			get
			{
				return m_rotation;
			}
			set
			{
				m_rotation = value;
			}
		}

		/// <summary>
		/// Position in coords
		/// </summary>
		public K3D.Geometry.Vector Position
		{
			get
			{
				return m_position;
			}
			set
			{
				m_position = value;
			}
		}

		/// <summary>
		/// Scaling factor
		/// </summary>
		public float Scaling
		{
			get
			{
				return m_scale;
			}
			set
			{
				m_scale = value;
			}
		}

		/// <summary>
		/// Rendering mode
		/// </summary>
		public DisplayMode DisplayMode
		{
			get
			{
				return m_displayMode;
			}
			set
			{
				m_displayMode = value;
				PropertiesService.Instance.DisplayMode = (int)m_displayMode;
				switch (m_displayMode)
				{
					case DisplayMode.Points:
                  Gl.glShadeModel(Gl.GL_FLAT);
						Gl.glDisable(Gl.GL_CULL_FACE);
                  Gl.glDisable(Gl.GL_LINE_SMOOTH);
						Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_POINT);
                  Transparency = Transparency;  // force transparency mode update
						break;
               case DisplayMode.AAWireframe:
                  Gl.glShadeModel(Gl.GL_FLAT);
                  Gl.glDisable(Gl.GL_CULL_FACE);
                  Gl.glEnable(Gl.GL_LINE_SMOOTH);
                  Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_LINE);
                  Gl.glEnable(Gl.GL_BLEND);     // force transparency on
                  break;
					case DisplayMode.Wireframe:
                  Gl.glShadeModel(Gl.GL_FLAT);
						Gl.glDisable(Gl.GL_CULL_FACE);
                  Gl.glDisable(Gl.GL_LINE_SMOOTH);
						Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_LINE);
                  Transparency = Transparency;  // force transparency mode update
						break;
					case DisplayMode.HiddenLine:
                  Gl.glShadeModel(Gl.GL_FLAT);
                  Gl.glEnable(Gl.GL_CULL_FACE);
                  Gl.glDisable(Gl.GL_LINE_SMOOTH);
                  Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);
                  Transparency = Transparency;  // force transparency mode update
						break;
					case DisplayMode.FlatShade:
                  Gl.glShadeModel(Gl.GL_FLAT);
						BackFaces = BackFaces;			// force backfaces mode update
                  Gl.glDisable(Gl.GL_LINE_SMOOTH);
						Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);
                  Transparency = Transparency;  // force transparency mode update
						break;
					case DisplayMode.SmoothShade:
                  Gl.glShadeModel(Gl.GL_SMOOTH);
						BackFaces = BackFaces;			// force backfaces mode update
                  Gl.glDisable(Gl.GL_LINE_SMOOTH);
						Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);
                  Transparency = Transparency;  // force transparency mode update
						break;
				}
			}
		}

      /// <summary>
      /// Toggle the display of rear facing polygons
      /// </summary>
      public bool BackFaces
      {
         get
         {
            return PropertiesService.Instance.BackFaces;
         }
         set
         {
            PropertiesService.Instance.BackFaces = value;
            if (value == true)
            {
               Gl.glDisable(Gl.GL_CULL_FACE);
            }
            else
            {
               Gl.glEnable(Gl.GL_CULL_FACE);
            }
         }
      }

      /// <summary>
      /// Transparency enabled
      /// </summary>
      public bool Transparency
      {
         get
         {
            return PropertiesService.Instance.Transparency;
         }
         set
         {
            if (value == true)
            {
               Gl.glEnable(Gl.GL_BLEND);
            }
            else
            {
               if (DisplayMode != DisplayMode.AAWireframe)
               {
                  Gl.glDisable(Gl.GL_BLEND);
               }
            }
            PropertiesService.Instance.Transparency = value;
         }
      }

      /// <summary>
      /// Background colour
      /// </summary>
      public System.Drawing.Color BackgroundColour
      {
         get
         {
            return PropertiesService.Instance.BackgroundColour;
         }
         set
         {
            PropertiesService.Instance.BackgroundColour = value;
            Gl.glClearColor( ((float)value.R)*ColourTriple.RGB_BYTE_TO_FLOAT,
               ((float)value.G)*ColourTriple.RGB_BYTE_TO_FLOAT,
               ((float)value.B)*ColourTriple.RGB_BYTE_TO_FLOAT, 0.0f );
         }
      }


      // ------------------------------------------------------------------------------
      // Private data

      /// <summary>
      /// Whether we should render the current 3D object during Paint event
      /// </summary>
      private bool m_bRender = true;

		/// <summary>
		/// OpenGL Display mode
		/// </summary>
		private DisplayMode m_displayMode = DisplayMode.SmoothShade;
		
      /// <summary>
      /// The 3D object to display
      /// </summary>
      private ThreeDObject m_3dObject = null;

      /// <summary>
      /// Active layer (on/off) list
      /// </summary>
      private ArrayList m_activeLayers = new ArrayList();
		
		/// <summary>
		/// Rotation, position and scaling factor
		/// </summary>
		private K3D.Geometry.Vector m_rotation = new K3D.Geometry.Vector(0, 180.0f, 0);
		private K3D.Geometry.Vector m_position = new K3D.Geometry.Vector();
		private float m_scale = 1.0f;

      /// <summary>
      /// Lighting defaults
      /// </summary>
      private static ColourTriple Light1Colour = new ColourTriple(1.0f, 1.0f, 1.0f);
      private static ColourTriple Light2Colour = new ColourTriple(0.95f, 0.95f, 1.0f);
      private static ColourTriple Light3Colour = new ColourTriple(0.95f, 1.0f, 0.95f);
      private static ColourTriple Light4Colour = new ColourTriple(1.0f, 1.0f, 1.0f);

      private const float Light1Intensity = 1.0f;
      private const float Light2Intensity = 0.3f;
      private const float Light3Intensity = 0.3f;
      private const float Light4Intensity = 0.3f;

      private const float Light1Specular = 0.7f;
      private const float Light2Specular = 0.3f;
      private const float Light3Specular = 0.5f;
      private const float Light4Specular = 0.1f;

		private static K3D.Geometry.Point Light1Position = new K3D.Geometry.Point(-10.0f, 15.0f, 100.0f);
      private static K3D.Geometry.Point Light2Position = new K3D.Geometry.Point(20.0f, 20.0f, 50.0f);
      private static K3D.Geometry.Point Light3Position = new K3D.Geometry.Point(25.0f, -25.0f, 150.0f);
      private static K3D.Geometry.Point Light4Position = new K3D.Geometry.Point(5.0f, 5.0f, -100.0f);
	}


	/// <summary>
	/// DisplayMode enum
	/// </summary>
	public enum DisplayMode
	{
		Points, Wireframe, AAWireframe, HiddenLine, FlatShade, SmoothShade
	};
}
