namespace K3D.Services
{
	using System;
	using System.IO;
	using System.Windows.Forms;
	using System.Collections;

	using Crownwood.Magic.Menus;


	/// <summary>
	/// MRU File List Service.
	/// </summary>
	public sealed class MRUFileListService
	{
		// --------------------------------------------------------------------------------
		// Construction

		/// <summary>
		/// Private constructor
		/// </summary>
		private MRUFileListService()
		{
			// get last MRU list from Registry
			string[] files = RegistryService.Instance.GetStringArray(REG_KEY, REG_NAME);
			if (files != null)
			{
				m_MRUlist.AddRange(files);
			}
		}
		
		/// <summary>
		/// Static instance
		/// </summary>
		public static MRUFileListService Instance
		{
			get
			{
				return s_service;
			}
		}


		// --------------------------------------------------------------------------------
		// Public API

		/// <summary>
		/// Set the parent menu for adding menu items too
		/// </summary>
		/// <param name="k3d">K3DGLForm</param>
		/// <param name="parent">MenuCommand</param>
		public void SetParentMenu(K3DGLForm k3d, MenuCommand parent)
		{
			if (k3d != null && parent != null)
			{
				m_k3dForm = k3d;
				m_parentMenu = parent;
				
				// build menus for the first time
				RebuildMenus();
			}
		}

		/// <summary>
		/// Add a new file (newest) to the MRU list
		/// </summary>
		/// <param name="filename">Full path of filename to add</param>
		public void AddFile(string filename)
		{
			if (filename != null && m_parentMenu != null)
			{
				// remove if already in the list at a previous position
				m_MRUlist.Remove(filename);

				// insert at top
				m_MRUlist.Insert(0, filename);

				// crop list to max length
				while (m_MRUlist.Count > MAX_LENGTH)
				{
					m_MRUlist.RemoveAt(MAX_LENGTH);
				}

				// save new list
				SaveMRUList();

				// rebuild menus
				RebuildMenus();
			}
		}

		/// <summary>
		/// Remove a file from the MRU list
		/// </summary>
		/// <param name="filename">Full path of the filename to remove</param>
		public void RemoveFile(string filename)
		{
			if (m_parentMenu != null)
			{
				m_MRUlist.Remove(filename);

				// save new list
				SaveMRUList();

				// rebuild menus
				RebuildMenus();
			}
		}


		// --------------------------------------------------------------------------------
		// Private helpers

		/// <summary>
		/// Rebuild the menus for the MRU list
		/// </summary>
		private void RebuildMenus()
		{
			m_parentMenu.MenuCommands.Clear();

			// init menus from current MRU list
			int index = 0;
			foreach (string file in m_MRUlist)
			{
				MenuCommand menu = new MenuCommand(ShortenPathname(file), new EventHandler(OnFileClick));
				menu.Tag = index++;			// tag with index into list
				m_parentMenu.MenuCommands.Add(menu);
			}
			
			m_parentMenu.Enabled = (m_MRUlist.Count != 0);
		}

		/// <summary>
		/// Persist the current MRU list to the registry
		/// </summary>
		private void SaveMRUList()
		{
			string[] files = (string[])m_MRUlist.ToArray(typeof(string));
			RegistryService.Instance.PutStringArray(REG_KEY, REG_NAME, files);
		}

		/// <summary>
		/// File menu click event
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnFileClick(object sender, EventArgs args)
		{
			MenuCommand menu = (MenuCommand)sender;
			m_k3dForm.LoadFile( (string)m_MRUlist[(int)menu.Tag], true );
		}

		// Shortens a pathname by either removing consecutive components of a path
		// and/or by removing characters from the end of the filename and replacing
		// then with three elipses (...)
		//
		// In all cases, the root of the passed path will be preserved in it's entirety.
		//
		// If a UNC path is used or the pathname and maxLength are particularly short,
		// the resulting path may be longer than maxLength.
		//
		// This method expects fully resolved pathnames to be passed to it.
		// (Use Path.GetFullPath() to obtain this.)
		public static String ShortenPathname(String pathname)
		{
			int maxLength = 64;

			if (pathname.Length <= maxLength)
				return pathname;

			String root = Path.GetPathRoot(pathname);
			if (root.Length > 3)
				root += Path.DirectorySeparatorChar;

			String[] elements = pathname.Substring(root.Length).Split(Path.DirectorySeparatorChar, Path.AltDirectorySeparatorChar);

			int filenameIndex = elements.GetLength(0) - 1;

			if (elements.GetLength(0) == 1) // pathname is just a root and filename
			{
				if (elements[0].Length > 5) // long enough to shorten
				{
					// if path is a UNC path, root may be rather long
					if (root.Length + 6 >= maxLength)
					{
						return root + elements[0].Substring(0, 3) + "...";
					}
					else
					{
						return pathname.Substring(0, maxLength - 3) + "...";
					}
				}
			}
			else if ((root.Length + 4 + elements[filenameIndex].Length) > maxLength) // pathname is just a root and filename
			{
				root += "...\\";

				int len = elements[filenameIndex].Length;
				if (len < 6)
					return root + elements[filenameIndex];

				if ((root.Length + 6) >= maxLength)
				{
					len = 3;
				}
				else
				{
					len = maxLength - root.Length - 3;
				}
				return root + elements[filenameIndex].Substring(0, len) + "...";
			}
			else if (elements.GetLength(0) == 2)
			{
				return root + "...\\" + elements[1];
			}
			else
			{
				int len = 0;
				int begin = 0;

				for (int i = 0; i < filenameIndex; i++)
				{
					if (elements[i].Length > len)
					{
						begin = i;
						len = elements[i].Length;
					}
				}

				int totalLength = pathname.Length - len + 3;
				int end = begin + 1;

				while (totalLength > maxLength)
				{
					if (begin > 0)
						totalLength -= elements[--begin].Length - 1;

					if (totalLength <= maxLength)
						break;

					if (end < filenameIndex)
						totalLength -= elements[++end].Length - 1;

					if (begin == 0 && end == filenameIndex)
						break;
				}

				// assemble final string
				for (int i = 0; i < begin; i++)
				{
					root += elements[i] + '\\';
				}

				root += "...\\";

				for (int i = end; i < filenameIndex; i++)
				{
					root += elements[i] + '\\';
				}

				return root + elements[filenameIndex];
			}
			return pathname;
		}


		// --------------------------------------------------------------------------------
		// Private data

		private const string REG_KEY	= "MRU";
		private const string REG_NAME	= "Files";
		private const int MAX_LENGTH	= 12;

		private static MRUFileListService s_service = new MRUFileListService();

		private ArrayList m_MRUlist = new ArrayList();

		private K3DGLForm m_k3dForm = null;
		private MenuCommand m_parentMenu = null;
	}
}
