namespace K3D.ThreeD.Lightwave
{
   using System;
   using System.Diagnostics;
   using System.Collections;

   using Tao.OpenGl;
   using Tao.Platform.Windows;

   using K3D.Geometry;
	using K3D.Services;


	/// <summary>
	/// LightwaveObject
	/// 
	/// Encapsulates all the data required to render a single Lightwave object.
	/// Support single and multi-layered object.
	/// </summary>
	public sealed class LightwaveObject : ThreeDObject
	{
		// ------------------------------------------------------------------------------
		// Public API

      /// <summary>
      /// Centre and scale this object
      /// </summary>
      public override void CentreAndScale()
      {
         // default size of object to scale too - calculate using initial window size etc.
         const float DEFAULTSIZE = 45.0f;

         // store for max object limits
         float minx = 0.0f, maxx = 0.0f;
         float miny = 0.0f, maxy = 0.0f;
         float minz = 0.0f, maxz = 0.0f;

         // walk all points to find object limits
         foreach (LightwaveLayer layer in m_layers)
         {
            Point[] points = layer.Points;
            for (int count=0; count<points.Length; count++) 
            {
               // x coord
               if (points[count].X < minx)
               {
                  minx = points[count].X;
               }
               else if (points[count].X > maxx)
               {
                  maxx = points[count].X;
               }
               
               // y coord
               if (points[count].Y < miny)
               {
                  miny = points[count].Y;
               }
               else if (points[count].Y > maxy)
               {
                  maxy = points[count].Y;
               }

               // z coord
               if (points[count].Z < minz)
               {
                  minz = points[count].Z;
               }
               else if (points[count].Z > maxz)
               {
                  maxz = points[count].Z;
               }
            }
         }

         // calculate centre displacement for object
         float cdispx = (minx + maxx)/2.0f;
         float cdispy = (miny + maxy)/2.0f;
         float cdispz = (minz + maxz)/2.0f;

         // subtract displacement from all points to centre object
         // get object points
			if (PropertiesService.Instance.AutoCentre == true)
			{
				foreach (LightwaveLayer layer in m_layers)
				{
					Point[] points = layer.Points;
					for (int count=0; count<points.Length; count++) 
					{
						points[count].X -= cdispx;
						points[count].Y -= cdispy;
						points[count].Z -= cdispz;
					}
				}
			}

         // calc total size multipliers using max object limits and scale
         float sizex = maxx - minx;
         float sizey = maxy - miny;
         float sizez = maxz - minz;
	
         // find largest of multipliers and use it as scale factor
         float scalefactor = 0.0f;
         if (sizey > sizex) 
         {
            if (sizez > sizey) 
            {
               // use sizez
               scalefactor = 1.0f / (sizez/DEFAULTSIZE);
            }
            else
            {
               // use sizey
               scalefactor = 1.0f / (sizey/DEFAULTSIZE);
            }
         }
         else if (sizez > sizex) 
         {
            // use sizez
            scalefactor = 1.0f / (sizez/DEFAULTSIZE);
         }
         else 
         {
            // use sizex
            scalefactor = 1.0f / (sizex/DEFAULTSIZE);
         }

         // multiply (scale) all points by size multipliers
         // get object points
			if (PropertiesService.Instance.AutoScale == true)
			{
				foreach (LightwaveLayer layer in m_layers)
				{
					Point[] points = layer.Points;
					for (int count=0; count<points.Length; count++) 
					{
						points[count].X *= scalefactor;
						points[count].Y *= scalefactor;
						points[count].Z *= scalefactor;
					}
				}
			}
      }

      /// <summary>
      /// Calculate face (facet) normals for a 3D object
      /// Normals are inverted to account for reversed Z coords
      /// </summary>
      public override void CalculateFacetNormals()
      {
         Point vertex1;
         Point vertex2;
         Point vertex3;
         
         // Lightwave uses the 1st, 2nd and final points in a poly to calc the normal vector
	      
         // polyindex used as index into polygon data
         foreach (LightwaveLayer layer in m_layers)
         {
            for (int polyCount=0; polyCount<layer.Polygons.Length; polyCount++)
            {
               // get number of points in the current polygon
               Polygon polygon = layer.Polygons[polyCount];
		         
               // skip single and 2 point polys
               if (polygon.VertexCount > 2) 
               {
                  // get vertices for a poly
                  // first vertex
                  vertex3 = layer.Points[ polygon.Vertices[0] ];
                  vertex2 = layer.Points[ polygon.Vertices[1] ];
                  vertex1 = layer.Points[ polygon.Vertices[polygon.VertexCount - 1] ];
			         
                  // calc and store normal for this poly
                  polygon.Normal = TriangleNormal(vertex1, vertex2, vertex3);
               }
            }
         }
      }

      /// <summary>
      /// Calculate averaged polygon normals for a 3D object.
      /// This allows gouraud shading to be applied to multi-vector polygons.
      /// For each finding all the polys that each point is referenced in,
      /// build a list of those references and average the normals
      /// </summary>
      public override void CalculateVertexNormals()
      {
         foreach (LightwaveLayer layer in m_layers)
         {
            int totalPoints = layer.Points.Length;
            int totalPolys  = layer.Polygons.Length;
            
            // this holds a list of poly indexes for each point in the layer
            UsedInPolySurfaceCollection[] usedLists = new UsedInPolySurfaceCollection[totalPoints];
            for (int polyIndex=0; polyIndex<totalPolys; polyIndex++)
            {
               Polygon polygon = layer.Polygons[polyIndex];

               if (polygon.VertexCount > 2 && polygon.Ignore == false)
               {
                  // walk the vertex index list in each poly and add this polygon
                  // index to the list for the specific vertex
                  for (int vertexIndex=0; vertexIndex<polygon.VertexCount; vertexIndex++)
                  {
                     int vertex = polygon.Vertices[vertexIndex];
                     
                     // create list for this point if none added yet
                     if (usedLists[vertex] == null)
                     {
                        usedLists[vertex] = new UsedInPolySurfaceCollection();
                     }
                     
                     // add the index of this poly to the list for this point
                     usedLists[vertex].AddUsedIn(polyIndex, polygon.SurfaceIndex);
                  }
               }
            }
            
            // for each of the vertices in the face, calculate the average
            // normal for that point by averaging the facet normals for all the
            // polys with the same surface it is used in and store the averaged
            // normal vector for that point
            Point normalA = new Point(0.0f, 0.0f, 0.0f);
            Point normalB = new Point(0.0f, 0.0f, 0.0f);
            for (int polyIndex=0; polyIndex<totalPolys; polyIndex++)
            {
               Polygon polygon = layer.Polygons[polyIndex];

               if (polygon.VertexCount > 2 && polygon.Ignore == false)
               {
                  Point[] normals = new Point[polygon.VertexCount];
                  Point average;
                  
                  // get max smooth angle from the surface for this polygon
						if (polygon.SurfaceIndex >= m_surfaces.Length)
						{
							if (K3DGLForm.DEBUG)	Trace.WriteLine("WARNING: Found unknown surface index: " + polygon.SurfaceIndex);
							polygon.SurfaceIndex = 0;
						}
                  LightwaveSurface surface = m_surfaces[polygon.SurfaceIndex];
                  double maxSmoothingAngle = surface.MaxSmoothing;	// radians
                  // calc smooth threshold
						double smoothThreshold = Math.Cos(maxSmoothingAngle);
                  
                  for (int vertexCount=0; vertexCount<polygon.VertexCount; vertexCount++)
                  {
                     int vertexIndex = polygon.Vertices[vertexCount];
                     
                     UsedInPolySurfaceCollection usedInList = usedLists[vertexIndex];
                     if (usedInList == null)
                     {
                        throw new Exception("Unable to find 'UsedInPolySurfaceCollection' for vertex index: " + vertexIndex);
                     }
                     
                     // start average with first normalised polygon normal
                     average = new Point( polygon.Normal );
							average = Normalize(average);

							// this is the poly normal we will calc each angle from
                     normalA.CopyFrom(average);
                     
                     // get all polys with the same surface used by this point
                     int[] polysWithSurface = usedInList.GetPolysWithSurface(polygon.SurfaceIndex);
                     for (int usedCount=0; usedCount<polysWithSurface.Length; usedCount++)
                     {
                        int polyUsedIndex = polysWithSurface[usedCount];
                        if (polyUsedIndex != polyIndex)
                        {
									// get the face normal for this poly and normalise it
                           normalB.CopyFrom( layer.Polygons[polyUsedIndex].Normal );
									normalB = Normalize(normalB);
									
                           // use cosine dot product and smoothing threshold angle to
                           // decide whether or not to average a polygon with its neighbour
                           if (Math.Abs(DotProduct(normalA, normalB)) >= smoothThreshold)
                           {
                              average.X += normalB.X;
                              average.Y += normalB.Y;
                              average.Z += normalB.Z;
                           }
                        }
                     }
                     
                     // normalise average vector
                     //Normalize(average);
                     
                     // save normal vector for this point
                     normals[vertexCount] = average;
                  }
                  
                  // save the vertex normals for this polygon
                  polygon.VertexNormals = normals;
               }
            }
         }
      }

      /// <summary>
      /// Convert the object data to OpenGL display lists
      /// </summary>
      public override void BuildOpenGLDisplayLists()
      {
         // material data
	      float[] material_ambient = {0.0f, 0.0f, 0.0f, 1.0f};
	      float[] material_diffuse = {1.0f, 1.0f, 1.0f, 1.0f};
	      float[] material_specular = {0.0f, 0.0f, 0.0f, 1.0f};
	      float[] material_shininess = {32.0f, 32.0f, 32.0f};   // max 128? - TODO: convert from float!
	      float[] material_emission = {0.0f, 0.0f, 0.0f, 1.0f}; // note: -1.0 to 1.0!
	      float[] default_colour = {0.75f, 0.75f, 0.75f};
         float transmission = 0.0f;

			double[] tessVert = new double[3];

			// cache properties for performance
			bool bShowPoints = PropertiesService.Instance.ShowPoints;
			bool bShowLines  = PropertiesService.Instance.ShowLines;
			bool bShowTris   = PropertiesService.Instance.ShowTris;
			bool bShowQuads  = PropertiesService.Instance.ShowQuads;
			bool bShowPolys  = PropertiesService.Instance.ShowPolys;
			bool bTesselate  = PropertiesService.Instance.Tesselate;

         // create GL display lists for object layers
         Gl.glGenLists(m_layers.Length);

         // create GLU polygon tessallator object and setup callbacks
	      Glu.GLUtesselator tess = Glu.gluNewTess();
         
         Glu.TessBeginCallback tessBeginProc = new Glu.TessBeginCallback(OnTessBegin);
         Glu.TessVertexCallback tessVertexProc = new Glu.TessVertexCallback(OnTessVertex);
         Glu.TessEndCallback tessEndProc = new Glu.TessEndCallback(OnTessEnd);

         Glu.gluTessCallback(tess, Glu.GLU_TESS_BEGIN, tessBeginProc);
         Glu.gluTessCallback(tess, Glu.GLU_TESS_VERTEX, tessVertexProc);
         Glu.gluTessCallback(tess, Glu.GLU_TESS_END, tessEndProc);

         ////////////////////////////////////////////////////////////
         // walk each layer in turn, compiling the polygons in each layer
         for (int layerCount=0; layerCount<m_layers.Length; layerCount++)
         {
            LightwaveLayer layer = m_layers[layerCount];
            Gl.glNewList(layerCount + 1, Gl.GL_COMPILE);
            
            // init default surface
            Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_SPECULAR, material_specular);
            Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_SHININESS, material_shininess);
            Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_EMISSION, material_emission);
            Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_DIFFUSE, material_diffuse);
            Gl.glColor4f(0.75f, 0.75f, 0.75f, 1.0f);

            int lastSurfaceIndex = -1;

            ////////////////////////////////////////////////////////////
            // Being an OpenGL compilation of polygon data
            // NOTE: LWO2Visitor could sort polygon types for potentially improved
				//		   performance when compiling the OpenGL structures
				int maxPolys = layer.Polygons.Length;
            for (int polyCount=0; polyCount<maxPolys; polyCount++)
            {
               Polygon polygon = layer.Polygons[polyCount];
               int numVertices = polygon.VertexCount;
               
               bool concavePolygon = false;
               bool ignorePolygon = polygon.Ignore;
               
               if (ignorePolygon == false)
               {
                  switch (numVertices)
                  {
                     case 1:
                        // single point
                        if (bShowPoints)
                           Gl.glBegin(Gl.GL_POINTS);
                        else
                           ignorePolygon = true;
                        break;
                     case 2:
                        // polyline (2 points)
                        if (bShowLines)
                           Gl.glBegin(Gl.GL_LINES);
                        else
                           ignorePolygon = true;
                        break;
                     case 3:
                        // generate triangle polygon
                        if (bShowTris)
                           Gl.glBegin(Gl.GL_TRIANGLES);
                        else
                           ignorePolygon = true;
                        break;
                     case 4:
                        // generate quad polygon
                        if (bShowQuads)
                           Gl.glBegin(Gl.GL_QUADS);
                        else
                           ignorePolygon = true;
                        break;
                     default:
                        // generate arbitary polygon
                        if (bShowPolys)
                        {
                           // if it's a concave polygon then use the GLU functions to tessellate it later
                           if (bTesselate)
                           {
                              concavePolygon = IsConcave(polygon, layer.Points);
                           }
                           else
                           {
                              concavePolygon = false;
                           }
                           if (concavePolygon == false)
                           {
                              Gl.glBegin(Gl.GL_POLYGON);
                           }
                        }
                        else
                        {
                           ignorePolygon = true;
                        }
                        break;
                  }
               }
               
               ////////////////////////////////////////////////////////////
               // Generate surface characteristics
               if (ignorePolygon == false)
               {
                  if (polygon.SurfaceIndex != lastSurfaceIndex) 
                  {
                     lastSurfaceIndex = polygon.SurfaceIndex;

							if (polygon.SurfaceIndex >= m_surfaces.Length)
							{
								if (K3DGLForm.DEBUG)	Trace.WriteLine("WARNING: Found unknown surface index: " + polygon.SurfaceIndex);
								polygon.SurfaceIndex = 0;
							}
                     LightwaveSurface surface = m_surfaces[polygon.SurfaceIndex];
                     if (true)//_theProperties.ReadSurfaces)
                     {
                        // colour diffusion
                        material_diffuse[0] = surface.Colour.Red * surface.Diffusion;
                        material_diffuse[1] = surface.Colour.Green * surface.Diffusion;
                        material_diffuse[2] = surface.Colour.Blue * surface.Diffusion;

                        // specular
                        // TODO: support colour highlights using flags and specular
                        //		   value as a multiplier against the colour value?
                        material_specular[0] = surface.Specular;
                        material_specular[1] = surface.Specular;
                        material_specular[2] = surface.Specular;

                        // emission (from luminosity)
                        material_emission[0] = surface.Luminosity;
                        material_emission[1] = surface.Luminosity;
                        material_emission[2] = surface.Luminosity;

                        // shininess (from gloss)
                        material_shininess[0] = surface.Glossiness * 128.0f;
                        material_shininess[1] = surface.Glossiness * 128.0f;
                        material_shininess[2] = surface.Glossiness * 128.0f;

                        // apply material
                        if (true)//_theProperties.ReadSpecular)
                        {
                           Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_SPECULAR, material_specular);
                           Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_SHININESS, material_shininess);
                        }
                        if (true)//_theProperties.ReadLuminosity)
                        {
                           Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_EMISSION, material_emission);
                        }
                        if (true)//_theProperties.ReadDiffusion)
                        {
									Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_DIFFUSE, material_diffuse);
                        }

                        // set transmission
                        if (true)//_theProperties.ReadTransparency)
                        {
                           transmission = surface.Transparency;
                        }
                        else
                        {
                           transmission = 0.0f;
                        }
                        // set colour
                        if (true)//_theProperties.ReadColour)
                        {
                           Gl.glColor4f(surface.Colour.Red * surface.Diffusion,
                              surface.Colour.Green * surface.Diffusion,
                              surface.Colour.Blue * surface.Diffusion,
                              1.0f - transmission);
                        }
                        else
                        {
                           Gl.glColor4f(default_colour[0],
                              default_colour[1],
                              default_colour[2],
                              1.0f - transmission);
                        }
                     }
                  }

                  ////////////////////////////////////////////////////////////
                  // Conversion of polygon to OpenGL

			         // if concave polygon then use the GLU tessellator functions
			         // to convert concave polygon to a series of convex polygons
			         if (concavePolygon)
			         {
							////////////////////////////////////////////////////////////
							// Concave polygon - tesselate into triangles
							m_currentTessLayer = layer;	// used during callback method
							
							LightwaveSurface surface = m_surfaces[polygon.SurfaceIndex];
							if (surface.Smoothing == true || polygon.ForceSmoothing == true)
							{
								//////////////////////////////////////////////////////////////////////////////////////
								// smooth concave poly
								Glu.gluTessBeginPolygon(tess, System.IntPtr.Zero);
					         for (int i=0; i<numVertices; i++)
					         {
									int vertexIndex = polygon.Vertices[i];
									
									Point normal = polygon.VertexNormals[i];
						         Glu.gluTessNormal(tess, -normal.X, -normal.Y, -normal.Z);
									
									Point vertex = layer.Points[ vertexIndex ];
									tessVert[0] = vertex.X;
									tessVert[1] = vertex.Y;
									tessVert[2] = vertex.Z;
         						
						         Glu.gluTessVertex(tess, tessVert, new System.IntPtr(vertexIndex));
					         }
								Glu.gluTessEndPolygon(tess);

								// render other side for double sided polys
								if (surface.DoubleSided == true)
								{
									Glu.gluTessBeginPolygon(tess, System.IntPtr.Zero);
									for (int i=numVertices-1; i>=0; i--)
									{
										int vertexIndex = polygon.Vertices[i];
										
										Point normal = polygon.VertexNormals[i];
										Glu.gluTessNormal(tess, normal.X, normal.Y, normal.Z);
										
										Point vertex = layer.Points[ vertexIndex ];
										tessVert[0] = vertex.X;
										tessVert[1] = vertex.Y;
										tessVert[2] = vertex.Z;
         							
										Glu.gluTessVertex(tess, tessVert, new System.IntPtr(vertexIndex));
									}
									Glu.gluTessEndPolygon(tess);
								}
				         }
				         else
				         {
								//////////////////////////////////////////////////////////////////////////////////////
								// flat concave poly
								Glu.gluTessBeginPolygon(tess, System.IntPtr.Zero);

                        // set face normal
					         Glu.gluTessNormal(tess, -polygon.Normal.X, -polygon.Normal.Y, -polygon.Normal.Z);
         					
					         for (int i=0; i<numVertices; i++)
					         {
									int vertexIndex = polygon.Vertices[i];
									
									Point vertex = layer.Points[ vertexIndex ];
									tessVert[0] = vertex.X;
									tessVert[1] = vertex.Y;
									tessVert[2] = vertex.Z;
									
						         Glu.gluTessVertex(tess, tessVert, new System.IntPtr(vertexIndex));
					         }
								Glu.gluTessEndPolygon(tess);

								// render other side for double sided polys
								if (surface.DoubleSided == true)
								{
									Glu.gluTessBeginPolygon(tess, System.IntPtr.Zero);

									Glu.gluTessNormal(tess, polygon.Normal.X, polygon.Normal.Y, polygon.Normal.Z);

									for (int i=numVertices-1; i>=0; i--)
									{
										int vertexIndex = polygon.Vertices[i];
										
										Point vertex = layer.Points[ vertexIndex ];
										tessVert[0] = vertex.X;
										tessVert[1] = vertex.Y;
										tessVert[2] = vertex.Z;
										
										Glu.gluTessVertex(tess, tessVert, new System.IntPtr(vertexIndex));
									}
									Glu.gluTessEndPolygon(tess);
								}
				         }
			         }
			         else
			         {
				         //////////////////////////////////////////////////////////////////////////////////////
				         // Convex polygon

				         // compile triangle vertices and normals
				         // set each normal for the vertex before each vertex
				         // if smoothing angle > 0 then use pointnormals, else use facetnormals
                     LightwaveSurface surface = m_surfaces[polygon.SurfaceIndex];
				         if (surface.Smoothing == true || polygon.ForceSmoothing == true)
				         {
								//////////////////////////////////////////////////////////////////////////////////////
								// smooth convex poly
					         for (int i=numVertices-1; i>=0; i--)
					         {
                           int vertexIndex = polygon.Vertices[i];
						         if (numVertices > 2)
						         {
										// TODO: normals here should be recalculated from new vertex data!?
                              // get the point normal for each vertex
                              Point normal = polygon.VertexNormals[i];
                              Gl.glNormal3f( normal.X, normal.Y, normal.Z );
						         }
                           
                           Point vertex = layer.Points[ vertexIndex ];
						         Gl.glVertex3f(vertex.X, vertex.Y, vertex.Z);
					         }

								// render other side for double sided polys
								if (surface.DoubleSided == true && numVertices > 2)
								{
									for (int i=0; i<numVertices; i++)
									{
										int vertexIndex = polygon.Vertices[i];
										if (numVertices > 2)
										{
											// get the point normal for each vertex
											Point normal = polygon.VertexNormals[i];
											Gl.glNormal3f(normal.X, normal.Y, normal.Z);
										}
                           
										Point vertex = layer.Points[ vertexIndex ];
										Gl.glVertex3f(vertex.X, vertex.Y, vertex.Z);
									}
								}
				         }
				         else
				         {
								//////////////////////////////////////////////////////////////////////////////////////
								// flat convex poly
					         if (numVertices > 2)
					         {
                           // set the single face normal for the polygon
						         Gl.glNormal3f(polygon.Normal.X, polygon.Normal.Y, polygon.Normal.Z);
					         }
					         for (int i=numVertices-1; i>=0; i--)
					         {
                           int vertexIndex = polygon.Vertices[i];
                           Point vertex = layer.Points[ vertexIndex ];
						         Gl.glVertex3f(vertex.X, vertex.Y, vertex.Z);
					         }

								// render other side for double sided polys
								if (surface.DoubleSided == true && numVertices > 2)
								{
									// get a single face normal for the polygon
									Gl.glNormal3f(polygon.Normal.X, polygon.Normal.Y, polygon.Normal.Z);

									for (int i=0; i<numVertices; i++)
									{
										int vertexIndex = polygon.Vertices[i];
										Point vertex = layer.Points[ vertexIndex ];
										Gl.glVertex3f(vertex.X, vertex.Y, vertex.Z);
									}
								}
				         }
				         // end polygon compilation
				         Gl.glEnd();
			         }
               }
            }
				// end display list complilation
				Gl.glEndList();
         }
			
         // delete tessellator object
         Glu.gluDeleteTess(tess);
         
         m_openGLlistLength = m_layers.Length;
      }

      /// <summary>
      /// Free the geometry structure memory
      /// </summary>
      public override void FreeGeometry()
      {
         m_layers = new LightwaveLayer[0];
      }

		/// <summary>
		/// GLUTesselator callback
		/// </summary>
		/// <param name="val"></param>
      private void OnTessBegin(Int32 val)
      {
         Gl.glBegin(val);
      }

		/// <summary>
		/// GLUTesselator callback
		/// </summary>
		/// <param name="ptr"></param>
      private void OnTessVertex(System.IntPtr ptr)
      {
         int vertexIndex = ptr.ToInt32();
         Gl.glVertex3f(m_currentTessLayer.Points[vertexIndex].X,
							  m_currentTessLayer.Points[vertexIndex].Y,
							  m_currentTessLayer.Points[vertexIndex].Z);
      }

		/// <summary>
		/// GLUTesselator callback
		/// </summary>
      private void OnTessEnd()
      {
         Gl.glEnd();
      }

      /// <summary>
      /// Return the number of OpenGL display list this object requires
      /// </summary>
      public override int OpenGLListLength
      {
         get
         {
            return m_openGLlistLength;
         }
      }

		/// <summary>
		/// The tag strings for this object
		/// </summary>
		public string[] Tags
		{
			get
			{
				return m_tags;
			}
			set
			{
				if (value != null)
				{
					m_tags = value;
				}
			}
		}

      /// <summary>
      /// The Lightwave Layers for this object
      /// </summary>
      public LightwaveLayer[] Layers
      {
         get
         {
            return m_layers;
         }
         set
         {
            if (value != null)
            {
               m_layers = value;
            }
         }
      }

      /// <summary>
      /// The surfaces for this object
      /// </summary>
      public LightwaveSurface[] Surfaces
      {
         get
         {
            return m_surfaces;
         }
         set
         {
            if (value != null)
            {
               m_surfaces = value;
            }
         }
      }

      public override string ToString()
      {
         string output = "LightwaveObject with " + m_surfaces.Length + " surfaces and " + m_layers.Length + " Layers\n";
         for (int count=0; count<m_layers.Length; count++)
         {
            output += "Layer: " + (count+1);
            output += "\n  " + m_layers[count].Points.Length + " points";
            output += "\n  " + m_layers[count].Polygons.Length + " polys\n";
				
				// TEMP
				//output += m_layers[count].ToString();
         }
         return output;
      }
      

		// ------------------------------------------------------------------------------
		// Private data

		/// <summary>
		/// The tag strings for this object
		/// </summary>
		private string[] m_tags = new string[0];

      /// <summary>
      /// Number of OpenGL lists in this object
      /// </summary>
      private int m_openGLlistLength = 0;

      /// <summary>
      /// The Lightwave Layers for this object
      /// </summary>
      private LightwaveLayer[] m_layers = new LightwaveLayer[0];

		/// <summary>
		/// Transient current layer used during tesselation callback
		/// </summary>
		private LightwaveLayer m_currentTessLayer = null;

      /// <summary>
      /// The surfaces for this object
      /// </summary>
      private LightwaveSurface[] m_surfaces = new LightwaveSurface[0];
	}
}
